"""
This file manages common global-level imports for which we want to centralize error handling
"""
import logging

logger = logging.getLogger(__name__)
sa_import_warning_required = False
spark_import_warning_required = False

try:
    import sqlalchemy as sa
except ImportError:
    logger.debug("No SqlAlchemy module available.")
    sa = None

try:
    from sqlalchemy.engine import Engine as sqlalchemy_engine_Engine
    from sqlalchemy.engine import Row as sqlalchemy_engine_Row
    from sqlalchemy.engine import reflection
except ImportError:
    logger.debug("No SqlAlchemy.engine module available.")
    reflection = None
    sqlalchemy_engine_Engine = None
    sqlalchemy_engine_Row = None

try:
    import sqlalchemy.func.count as sa_func_count
except ImportError:
    logger.debug("No SqlAlchemy.func module available.")
    sa_func_count = None

try:
    import sqlalchemy.sql.expression.ColumnClause as sa_sql_expression_ColumnClause
    import sqlalchemy.sql.expression.Select as sa_sql_expression_Select
    import sqlalchemy.sql.expression.Selectable as sa_sql_expression_Selectable
except ImportError:
    logger.debug("No SqlAlchemy.sql.expression module available.")
    sa_sql_expression_ColumnClause = None
    sa_sql_expression_Select = None
    sa_sql_expression_Selectable = None

try:
    from sqlalchemy.sql.elements import quoted_name
except ImportError:
    logger.debug("No SqlAlchemy.sql.elements module available.")
    quoted_name = None

try:
    import pyspark.sql.functions as F
    import pyspark.sql.types as sparktypes
except ImportError:
    logger.debug("No spark functions module available.")
    sparktypes = None
    F = None

try:
    from pyspark.ml.feature import Bucketizer
except ImportError:
    logger.debug("No spark Bucketizer available.")
    Bucketizer = None

try:
    from pyspark.sql import Window
except ImportError:
    logger.debug("No spark Window function available.")
    Window = None

try:
    from pyspark.sql import Column as pyspark_sql_Column
    from pyspark.sql import DataFrame as pyspark_sql_DataFrame
    from pyspark.sql import Row as pyspark_sql_Row
    from pyspark.sql import SparkSession as pyspark_sql_SparkSession
    from pyspark.sql import SQLContext
except ImportError:
    logger.debug("No spark SQLContext available.")
    SQLContext = None
    pyspark_sql_Column = None
    pyspark_sql_DataFrame = None
    pyspark_sql_Row = None
    pyspark_sql_SparkSession = None



import pytest
from moto import mock_glue
from ruamel.yaml import YAML

from great_expectations import DataContext
from great_expectations.core.batch import Batch, BatchDefinition, BatchRequest
from great_expectations.core.id_dict import IDDict
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.datasource import Datasource
from great_expectations.datasource.data_connector import (
    ConfiguredAssetAWSGlueDataCatalogDataConnector,
)
from great_expectations.exceptions import DataConnectorError
from great_expectations.validator.validator import Validator

yaml = YAML(typ="safe")


def test_instantiation_with_partitions_in_connector(glue_titanic_catalog):
    config = yaml.load(
        """
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME
    partitions:
        - PClass
        - SexCode

    assets:
        db_test.tb_titanic:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            splitter_method: split_on_column_value
            splitter_kwargs:
                column_name: PClass
            partitions:
                - SexCode
        asset2:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            sampling_method: _sample_using_random
            sampling_kwargs:
                p: 0.5
                seed: 0
            partitions:
                - PClass
        asset3:
            table_name: tb_titanic_with_partitions
            database_name: db_test
    """,
    )
    config["execution_engine"] = "execution_engine"

    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(**config)

    report = my_data_connector.self_check()
    assert report == {
        "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
        "data_asset_count": 3,
        "example_data_asset_names": ["asset2", "asset3", "db_test.tb_titanic"],
        "data_assets": {
            "asset2": {
                "batch_definition_count": 3,
                "example_data_references": [
                    {"PClass": "1st"},
                    {"PClass": "2nd"},
                    {"PClass": "3rd"},
                ],
            },
            "asset3": {
                "batch_definition_count": 6,
                "example_data_references": [
                    {"PClass": "1st", "SexCode": "0"},
                    {"PClass": "1st", "SexCode": "1"},
                    {"PClass": "2nd", "SexCode": "0"},
                ],
            },
            "db_test.tb_titanic": {
                "batch_definition_count": 2,
                "example_data_references": [
                    {"SexCode": "0"},
                    {"SexCode": "1"},
                ],
            },
        },
        "unmatched_data_reference_count": 0,
        "example_unmatched_data_references": [],
    }
    # noinspection PyProtectedMember
    assert len(my_data_connector.get_available_data_asset_names()) == 3
    assert my_data_connector.get_unmatched_data_references() == []


def test_basic_instantiation(glue_titanic_catalog):
    config = yaml.load(
        """
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME

    assets:
        db_test.tb_titanic:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            splitter_method: split_on_column_value
            splitter_kwargs:
                column_name: PClass
        asset2:
            table_name: tb_titanic_without_partitions
            database_name: db_test
            sampling_method: _sample_using_random
            sampling_kwargs:
                p: 0.5
                seed: 0
            partitions:
                - PClass
        asset3:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            partitions:
                - PClass
        asset4:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            partitions:
                - PClass
                - SexCode
    """,
    )
    config["execution_engine"] = "execution_engine"

    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(**config)

    report = my_data_connector.self_check()
    assert report == {
        "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
        "data_asset_count": 4,
        "example_data_asset_names": ["asset2", "asset3", "asset4"],
        "data_assets": {
            "asset2": {
                "batch_definition_count": 1,
                "example_data_references": [{}],
            },
            "asset3": {
                "batch_definition_count": 3,
                "example_data_references": [
                    {"PClass": "1st"},
                    {"PClass": "2nd"},
                    {"PClass": "3rd"},
                ],
            },
            "asset4": {
                "batch_definition_count": 6,
                "example_data_references": [
                    {"PClass": "1st", "SexCode": "0"},
                    {"PClass": "1st", "SexCode": "1"},
                    {"PClass": "2nd", "SexCode": "0"},
                ],
            },
        },
        "unmatched_data_reference_count": 0,
        "example_unmatched_data_references": [],
    }
    # noinspection PyProtectedMember
    assert len(my_data_connector.get_available_data_asset_names()) == 4
    assert my_data_connector.get_unmatched_data_references() == []


def test_invalid_instantiation(glue_titanic_catalog):
    config = yaml.load(
        """
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME

    assets:
        asset1:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            partitions:
                PClass
    """,
    )
    config["execution_engine"] = "execution_engine"
    with pytest.raises(DataConnectorError) as excinfo:
        ConfiguredAssetAWSGlueDataCatalogDataConnector(**config)
    assert "partitions" in str(excinfo.value)

    config = yaml.load(
        """
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME

    assets:
        asset1:
            table_name: tb_titanic_with_partitions
            partitions:
                - PClass
    """,
    )
    config["execution_engine"] = "execution_engine"

    with pytest.raises(DataConnectorError) as excinfo:
        ConfiguredAssetAWSGlueDataCatalogDataConnector(**config)
    assert "database_name" in str(excinfo.value)

    config = yaml.load(
        """
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME

    assets:
        asset1:
            database_name: db_test
            partitions:
                - PClass
    """,
    )
    config["execution_engine"] = "execution_engine"

    with pytest.raises(DataConnectorError) as excinfo:
        ConfiguredAssetAWSGlueDataCatalogDataConnector(**config)
    assert "table_name" in str(excinfo.value)


def test_instantiation_from_a_config(
    empty_data_context_stats_enabled, glue_titanic_catalog
):
    context: DataContext = empty_data_context_stats_enabled
    report_object = context.test_yaml_config(
        """
    module_name: great_expectations.datasource.data_connector
    class_name: ConfiguredAssetAWSGlueDataCatalogDataConnector
    name: my_glue_catalog_data_connector
    datasource_name: FAKE_Datasource_NAME

    assets:
        db_test.tb_titanic:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            splitter_method: split_on_column_value
            splitter_kwargs:
                column_name: PClass
        asset2:
            table_name: tb_titanic_without_partitions
            database_name: db_test
            sampling_method: _sample_using_random
            sampling_kwargs:
                p: 0.5
                seed: 0
        asset3:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            partitions:
                - PClass
        asset4:
            table_name: tb_titanic_with_partitions
            database_name: db_test
            partitions:
                - PClass
                - SexCode
    """,
        runtime_environment={
            "execution_engine": "execution_engine",
        },
        return_mode="report_object",
    )

    assert report_object == {
        "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
        "data_asset_count": 4,
        "example_data_asset_names": ["asset2", "asset3", "asset4"],
        "data_assets": {
            "asset2": {
                "batch_definition_count": 1,
                "example_data_references": [{}],
            },
            "asset3": {
                "batch_definition_count": 3,
                "example_data_references": [
                    {"PClass": "1st"},
                    {"PClass": "2nd"},
                    {"PClass": "3rd"},
                ],
            },
            "asset4": {
                "batch_definition_count": 6,
                "example_data_references": [
                    {"PClass": "1st", "SexCode": "0"},
                    {"PClass": "1st", "SexCode": "1"},
                    {"PClass": "2nd", "SexCode": "0"},
                ],
            },
        },
        "unmatched_data_reference_count": 0,
        "example_unmatched_data_references": [],
    }


def test_get_batch_definition_list_from_batch_request(glue_titanic_catalog):
    my_data_connector_yaml = yaml.load(
        f"""
        module_name: great_expectations.datasource.data_connector
        class_name: ConfiguredAssetAWSGlueDataCatalogDataConnector
        datasource_name: FAKE_Datasource_NAME
        name: my_glue_catalog_data_connector
        assets:
            with_partitions:
                database_name: db_test
                table_name: tb_titanic_with_partitions
                partitions:
                    - PClass
                    - SexCode
            without_partitions:
                database_name: db_test
                table_name: tb_titanic_with_partitions
        """,
    )
    my_data_connector: ConfiguredAssetAWSGlueDataCatalogDataConnector = (
        instantiate_class_from_config(
            config=my_data_connector_yaml,
            runtime_environment={
                "execution_engine": "execution_engine",
            },
            config_defaults={
                "module_name": "great_expectations.datasource.data_connector"
            },
        )
    )
    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="with_partitions",
                data_connector_query={"batch_filter_parameters": {"PClass": "1st"}},
            )
        )
    )
    assert len(batch_definition_list) == 2
    assert batch_definition_list[0]["batch_identifiers"] == {
        "PClass": "1st",
        "SexCode": "0",
    }
    assert batch_definition_list[1]["batch_identifiers"] == {
        "PClass": "1st",
        "SexCode": "1",
    }

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="with_partitions",
                data_connector_query={"batch_filter_parameters": {"SexCode": "0"}},
            )
        )
    )
    assert len(batch_definition_list) == 3
    assert batch_definition_list[0]["batch_identifiers"] == {
        "PClass": "1st",
        "SexCode": "0",
    }
    assert batch_definition_list[1]["batch_identifiers"] == {
        "PClass": "2nd",
        "SexCode": "0",
    }
    assert batch_definition_list[2]["batch_identifiers"] == {
        "PClass": "3rd",
        "SexCode": "0",
    }

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="with_partitions",
                data_connector_query={"batch_filter_parameters": {}},
            )
        )
    )
    assert len(batch_definition_list) == 6

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="with_partitions",
                data_connector_query={"batch_filter_parameters": {"SexCode": "1st"}},
            )
        )
    )
    assert len(batch_definition_list) == 0

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="with_partitions",
            )
        )
    )
    assert len(batch_definition_list) == 6

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="without_partitions",
                data_connector_query={"batch_filter_parameters": {"PClass": "1st"}},
            )
        )
    )
    assert len(batch_definition_list) == 0

    batch_definition_list = (
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="without_partitions",
            )
        )
    )
    assert len(batch_definition_list) == 1
    assert batch_definition_list[0]["batch_identifiers"] == {}

    with pytest.raises(KeyError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
                data_asset_name="db_test.fake",
            )
        )

    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(
                datasource_name="FAKE_Datasource_NAME",
                data_connector_name="my_glue_catalog_data_connector",
            )
        )

    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest(datasource_name="FAKE_Datasource_NAME")
        )

    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list_from_batch_request(
            batch_request=BatchRequest()
        )


@mock_glue
def test_instantiation_with_minimal_options():
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="execution_engine",
    )
    assert len(my_data_connector.assets) == 0


def test_instantiation_with_batch_spec_passthrough(
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine=execution_engine,
        assets={
            "titanic_asset": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
            },
        },
        batch_spec_passthrough={
            "splitter_method": "_split_on_whole_table",
            "splitter_kwargs": {},
            "sampling_method": "_sample_using_random",
            "sampling_kwargs": {"p": 0.5, "seed": 0},
        },
    )

    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=BatchDefinition(
            datasource_name="FAKE_Datasource_NAME",
            data_connector_name="my_glue_catalog_data_connector",
            data_asset_name="titanic_asset",
            batch_identifiers=IDDict(),
        )
    )

    batch = Batch(data=batch_data)

    validator = Validator(execution_engine, batches=[batch])
    # The sampling probability "p" used in "SparkDFExecutionEngine._sample_using_random()" is 0.5 (the equivalent of a
    # fair coin with the 50% chance of coming up as "heads").  Hence, on average we should get as much as 60% of the rows.
    assert len(validator.head(fetch_all=True)) < 788


def test_instantiation_with_asset_suffix_and_prefix(glue_titanic_catalog):
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="execution_engine",
        assets={
            "db_test.tb_titanic": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
                "data_asset_name_prefix": "prefix__",
                "data_asset_name_suffix": "__suffix",
            },
        },
    )
    assert "prefix__db_test.tb_titanic__suffix" in my_data_connector.assets


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_instantiation_with_splitter_sampling_and_prefix(
    splitter_method_name_prefix, glue_titanic_catalog
):
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="execution_engine",
        assets={
            "db_test.tb_titanic": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
                "splitter_method": f"{splitter_method_name_prefix}split_on_column_value",
                "splitter_kwargs": {
                    "column_name": "PClass",
                    "batch_identifiers": {"PClass": "1st"},
                },
                "sampling_method": "_sample_using_random",
                "sampling_kwargs": {"p": 0.5, "seed": 0},
                "data_asset_name_prefix": "prefix__",
                "data_asset_name_suffix": "__suffix",
            },
        },
    )

    data_asset_name = "prefix__db_test.tb_titanic__suffix"
    asset_config = my_data_connector.assets.get(data_asset_name, None)
    assert asset_config and asset_config == {
        "table_name": "tb_titanic_with_partitions",
        "database_name": "db_test",
        "splitter_method": f"{splitter_method_name_prefix}split_on_column_value",
        "splitter_kwargs": {
            "column_name": "PClass",
            "batch_identifiers": {"PClass": "1st"},
        },
        "sampling_method": "_sample_using_random",
        "sampling_kwargs": {"p": 0.5, "seed": 0},
        "data_asset_name_prefix": "prefix__",
        "data_asset_name_suffix": "__suffix",
    }


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_build_batch_spec_with_all_options(
    splitter_method_name_prefix, glue_titanic_catalog
):
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="execution_engine",
        assets={
            "asset_titanic": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
                "splitter_method": f"{splitter_method_name_prefix}split_on_column_value",
                "splitter_kwargs": {
                    "column_name": "PClass",
                    "batch_identifiers": {"PClass": "1st"},
                },
                "sampling_method": f"{splitter_method_name_prefix}sample_using_random",
                "sampling_kwargs": {"p": 0.5, "seed": 0},
                "data_asset_name_prefix": "prefix__",
                "data_asset_name_suffix": "__suffix",
            },
        },
    )
    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="prefix__asset_titanic__suffix",
        batch_identifiers=IDDict(),
    )

    batch_spec = my_data_connector.build_batch_spec(batch_definition)
    assert batch_spec.database_name == "db_test"
    assert batch_spec.table_name == "tb_titanic_with_partitions"
    assert batch_spec.get("data_asset_name", None) == "prefix__asset_titanic__suffix"
    assert (
        batch_spec.get("splitter_method", None)
        == f"{splitter_method_name_prefix}split_on_column_value"
    )
    assert batch_spec.get("splitter_kwargs", {}) == {
        "column_name": "PClass",
        "batch_identifiers": {"PClass": "1st"},
    }
    assert (
        batch_spec.get("sampling_method", None)
        == f"{splitter_method_name_prefix}sample_using_random"
    )
    assert batch_spec.get("sampling_kwargs", {}) == {"p": 0.5, "seed": 0}
    assert batch_spec.get("data_asset_name_prefix", None) == "prefix__"
    assert batch_spec.get("data_asset_name_suffix", None) == "__suffix"


def test_build_batch_spec_with_minimal_options(glue_titanic_catalog):
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine",
        assets={
            "db_test.tb_titanic": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
            },
        },
    )
    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
    )

    batch_spec = my_data_connector.build_batch_spec(batch_definition)
    assert batch_spec.database_name == "db_test"
    assert batch_spec.table_name == "tb_titanic_with_partitions"
    assert batch_spec.get("data_asset_name", None) == "db_test.tb_titanic"


def test_get_unmatched_data_references(glue_titanic_catalog):
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine",
        assets={
            "db_test.tb_titanic": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
            },
        },
    )
    unmatched_references = my_data_connector.get_unmatched_data_references()
    assert len(unmatched_references) == 0


def test_get_available_data_asset_names(glue_titanic_catalog):
    # noinspection PyTypeChecker
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine="test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine",
        assets={
            "asset1": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
            }
        },
    )
    data_asset_names = my_data_connector.get_available_data_asset_names()
    assert "asset1" in data_asset_names


def test_get_batch_data_and_metadata_with_sampling_method__random_in_asset_config(
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine=execution_engine,
        assets={
            "titanic_asset": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
                "splitter_method": "_split_on_whole_table",
                "splitter_kwargs": {},
                "sampling_method": "_sample_using_random",
                "sampling_kwargs": {"p": 0.5, "seed": 0},
            },
        },
    )

    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=BatchDefinition(
            datasource_name="FAKE_Datasource_NAME",
            data_connector_name="my_glue_catalog_data_connector",
            data_asset_name="titanic_asset",
            batch_identifiers=IDDict(),
            batch_spec_passthrough={},
        )
    )

    batch = Batch(data=batch_data)

    validator = Validator(execution_engine, batches=[batch])
    # The sampling probability "p" used in "SparkDFExecutionEngine._sample_using_random()" is 0.5 (the equivalent of a
    # fair coin with the 50% chance of coming up as "heads").  Hence, on average we should get as much as 60% of the rows.
    assert len(validator.head(fetch_all=True)) < 788


def test_get_batch_data_and_metadata_with_sampling_method__random_in_batch_spec_passthrough(
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    my_data_connector = ConfiguredAssetAWSGlueDataCatalogDataConnector(
        name="my_glue_catalog_data_connector",
        datasource_name="FAKE_Datasource_NAME",
        execution_engine=execution_engine,
        assets={
            "titanic_asset": {
                "table_name": "tb_titanic_with_partitions",
                "database_name": "db_test",
            },
        },
    )

    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=BatchDefinition(
            datasource_name="FAKE_Datasource_NAME",
            data_connector_name="my_glue_catalog_data_connector",
            data_asset_name="titanic_asset",
            batch_identifiers=IDDict(),
            batch_spec_passthrough={
                "splitter_method": "_split_on_whole_table",
                "splitter_kwargs": {},
                "sampling_method": "_sample_using_random",
                "sampling_kwargs": {"p": 0.5, "seed": 0},
            },
        )
    )

    batch = Batch(data=batch_data)

    validator = Validator(execution_engine, batches=[batch])
    # The sampling probability "p" used in "SparkDFExecutionEngine._sample_using_random()" is 0.5 (the equivalent of a
    # fair coin with the 50% chance of coming up as "heads").  Hence, on average we should get as much as 60% of the rows.
    assert len(validator.head(fetch_all=True)) < 788


def test_get_batch_data_and_metadata_with_sampling_method__mod(
    in_memory_runtime_context,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": "_split_on_whole_table",
            "splitter_kwargs": {},
            "sampling_method": "_sample_using_mod",
            "sampling_kwargs": {"column_name": "Age", "mod": 25, "value": 1},
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_not_be_in_set("Age", [25, 50]).success
    assert len(validator.head(fetch_all=True)) == 38


def test_get_batch_data_and_metadata_with_sampling_method__list(
    in_memory_runtime_context,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": "_split_on_whole_table",
            "splitter_kwargs": {},
            "sampling_method": "_sample_using_a_list",
            "sampling_kwargs": {"column_name": "PClass", "value_list": ["1st"]},
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set("PClass", ["1st"]).success
    assert len(validator.head(fetch_all=True)) == 322


def test_get_batch_data_and_metadata_with_sampling_method__hash(
    in_memory_runtime_context,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": "_split_on_whole_table",
            "splitter_kwargs": {},
            "sampling_method": "_sample_using_hash",
            "sampling_kwargs": {
                "column_name": "Name",
                "hash_function_name": "md5",
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert len(validator.head(fetch_all=True)) == 77


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__whole_table(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_whole_table",
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert len(validator.head(fetch_all=True)) == 1313


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__column_value(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_column_value",
            "splitter_kwargs": {
                "column_name": "PClass",
                "batch_identifiers": {"PClass": "1st"},
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set("PClass", ["1st"]).success
    assert len(validator.head(fetch_all=True)) == 322


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__divided_integer(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_divided_integer",
            "splitter_kwargs": {
                "column_name": "Age",
                "divisor": 20,
                "batch_identifiers": {"Age": 1},
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set(
        "Age", list(range(20, 40))
    ).success
    assert len(validator.head(fetch_all=True)) == 420


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__mod_integer(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_mod_integer",
            "splitter_kwargs": {
                "column_name": "Age",
                "mod": 25,
                "batch_identifiers": {"Age": 1},
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set("Age", [26, 1, 1.5, 51]).success
    assert len(validator.head(fetch_all=True)) == 38


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__multi_column_values(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_multi_column_values",
            "splitter_kwargs": {
                "column_names": ["Age", "PClass"],
                "batch_identifiers": {"Age": 25, "PClass": "1st"},
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set("Age", [25]).success
    assert validator.expect_column_values_to_be_in_set("PClass", ["1st"]).success
    assert len(validator.head(fetch_all=True)) == 4


@pytest.mark.parametrize("splitter_method_name_prefix", ["_", ""])
def test_get_batch_data_and_metadata_with_splitting_method__hashed_column(
    in_memory_runtime_context,
    splitter_method_name_prefix,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict(),
        batch_spec_passthrough={
            "splitter_method": f"{splitter_method_name_prefix}split_on_hashed_column",
            "splitter_kwargs": {
                "column_name": "Name",
                "hash_digits": 1,
                "hash_function_name": "md5",
                "batch_identifiers": {"hash_value": "f"},
            },
        },
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert len(validator.head(fetch_all=True)) == 77


def test_get_batch_data_and_metadata_with_partitions(
    in_memory_runtime_context,
    test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine,
    glue_titanic_catalog,
):
    execution_engine = (
        test_cases_for_aws_glue_data_catalog_data_connector_spark_execution_engine
    )
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"] = Datasource(
        name="FAKE_Datasource_NAME",
        # Configuration for "execution_engine" here is largely placeholder to comply with "Datasource" constructor.
        execution_engine=execution_engine.config,
        data_connectors={
            "my_glue_catalog_data_connector": {
                "class_name": "ConfiguredAssetAWSGlueDataCatalogDataConnector",
                "assets": {
                    "db_test.tb_titanic": {
                        "table_name": "tb_titanic_with_partitions",
                        "database_name": "db_test",
                        "partitions": ["PClass", "SexCode"],
                    },
                },
            },
        },
    )
    # Updating "execution_engine" to insure peculiarities, incorporated herein, propagate to "ExecutionEngine" itself.
    in_memory_runtime_context.datasources["FAKE_Datasource_NAME"]._execution_engine = execution_engine  # type: ignore[union-attr]

    batch_definition = BatchDefinition(
        datasource_name="FAKE_Datasource_NAME",
        data_connector_name="my_glue_catalog_data_connector",
        data_asset_name="db_test.tb_titanic",
        batch_identifiers=IDDict({"PClass": "2nd", "SexCode": "1"}),
    )
    my_data_connector = in_memory_runtime_context.datasources[
        "FAKE_Datasource_NAME"
    ].data_connectors["my_glue_catalog_data_connector"]
    batch_data, _, __ = my_data_connector.get_batch_data_and_metadata(
        batch_definition=batch_definition
    )
    batch = Batch(data=batch_data, batch_definition=batch_definition)
    validator = Validator(
        execution_engine, batches=[batch], data_context=in_memory_runtime_context
    )
    assert validator.expect_column_values_to_be_in_set("PClass", ["2nd"]).success
    assert validator.expect_column_values_to_be_in_set("SexCode", ["1"]).success



from __future__ import annotations

import copy
import datetime
import json
import logging
import sys
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.core._docs_decorators import public_api
from great_expectations.core.batch import (
    Batch,
    BatchRequestBase,
    batch_request_contains_batch_data,
)
from great_expectations.core.config_peer import ConfigPeer
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.usage_statistics import (
    UsageStatisticsHandler,
    get_profiler_run_usage_statistics,
    usage_statistics_enabled_method,
)
from great_expectations.core.util import (
    convert_to_json_serializable,
    determine_progress_bar_method_by_environment,
    nested_update,
)
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.types.refs import GXCloudResourceRef
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.rule_based_profiler import RuleBasedProfilerResult
from great_expectations.rule_based_profiler.config.base import (
    DomainBuilderConfig,
    ExpectationConfigurationBuilderConfig,
    ParameterBuilderConfig,
    RuleBasedProfilerConfig,
    domainBuilderConfigSchema,
    expectationConfigurationBuilderConfigSchema,
    parameterBuilderConfigSchema,
    ruleBasedProfilerConfigSchema,
)
from great_expectations.rule_based_profiler.expectation_configuration_builder import (
    ExpectationConfigurationBuilder,
    init_rule_expectation_configuration_builders,
)
from great_expectations.rule_based_profiler.helpers.configuration_reconciliation import (
    DEFAULT_RECONCILATION_DIRECTIVES,
    ReconciliationDirectives,
    ReconciliationStrategy,
    reconcile_rule_variables,
)
from great_expectations.rule_based_profiler.helpers.util import (
    convert_variables_to_dict,
)
from great_expectations.rule_based_profiler.parameter_builder import (
    ParameterBuilder,
    init_rule_parameter_builders,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,
    ParameterNode,
    build_parameter_container_for_variables,
)
from great_expectations.rule_based_profiler.rule import Rule, RuleOutput
from great_expectations.rule_based_profiler.rule.rule_state import RuleState
from great_expectations.util import filter_properties_dict

if TYPE_CHECKING:
    from great_expectations.core.domain import Domain
    from great_expectations.core.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.data_context.store.profiler_store import ProfilerStore
    from great_expectations.rule_based_profiler.domain_builder.domain_builder import (
        DomainBuilder,
    )
    from great_expectations.rule_based_profiler.helpers.runtime_environment import (
        RuntimeEnvironmentDomainTypeDirectives,
        RuntimeEnvironmentVariablesDirectives,
    )


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BaseRuleBasedProfiler(ConfigPeer):
    """
    BaseRuleBasedProfiler class is initialized from RuleBasedProfilerConfig typed object and contains all functionality
    in the form of interface methods (which can be overwritten by subclasses) and their reference implementation.
    """

    EXPECTATION_SUCCESS_KEYS: Set[str] = {
        "auto",
        "profiler_config",
    }

    def __init__(
        self,
        profiler_config: RuleBasedProfilerConfig,
        data_context: Optional[AbstractDataContext] = None,
        usage_statistics_handler: Optional[UsageStatisticsHandler] = None,
    ) -> None:
        """
        Create a new RuleBasedProfilerBase using configured rules (as captured in the RuleBasedProfilerConfig object).

        For a Rule or an item in a Rule configuration, instantiates the following if
        available: a domain builder, a parameter builder, and a configuration builder.
        These will be used to define profiler computation patterns.

        Args:
            profiler_config: RuleBasedProfilerConfig -- formal typed object containing configuration
            data_context: AbstractDataContext object that defines full runtime environment (data access, etc.)
        """
        name: str = profiler_config.name
        id: Optional[str] = None
        if hasattr(profiler_config, "id"):
            id = profiler_config.id
        config_version: float = profiler_config.config_version
        variables: Optional[Dict[str, Any]] = profiler_config.variables
        rules: Optional[Dict[str, Dict[str, Any]]] = profiler_config.rules

        self._name = name
        self._id = id
        self._config_version = config_version

        self._profiler_config = profiler_config

        if variables is None:
            variables = {}

        # Convert variables argument to ParameterContainer
        _variables: ParameterContainer = build_parameter_container_for_variables(
            variables_configs=variables
        )
        self.variables = _variables

        self._usage_statistics_handler = usage_statistics_handler

        self._data_context = data_context

        self._rules = self._init_profiler_rules(rules=rules)

        self._rule_states = []

    @property
    def ge_cloud_id(self) -> Optional[str]:
        return self._id

    @ge_cloud_id.setter
    def ge_cloud_id(self, id: str) -> None:
        self._id = id

    def _init_profiler_rules(
        self,
        rules: Dict[str, Dict[str, Any]],
    ) -> List[Rule]:
        if rules is None:
            rules = {}

        rule_object_list: List[Rule] = []

        rule_name: str
        rule_config: Dict[str, Any]
        for rule_name, rule_config in rules.items():
            rule_object_list.append(
                self._init_rule(rule_name=rule_name, rule_config=rule_config)
            )

        return rule_object_list

    def _init_rule(
        self,
        rule_name: str,
        rule_config: Dict[str, Any],
    ) -> Rule:
        # Config is validated through schema but do a sanity check
        attr: str
        for attr in (
            "domain_builder",
            "expectation_configuration_builders",
        ):
            if attr not in rule_config:
                raise gx_exceptions.ProfilerConfigurationError(
                    message=f'Invalid rule "{rule_name}": missing mandatory {attr}.'
                )

        # Instantiate variables and builder attributes
        variables: Dict[str, Any] = rule_config.get("variables", {})
        domain_builder: DomainBuilder = RuleBasedProfiler._init_rule_domain_builder(
            domain_builder_config=rule_config.get("domain_builder"),
            data_context=self._data_context,
        )
        parameter_builders: Optional[
            List[ParameterBuilder]
        ] = init_rule_parameter_builders(
            parameter_builder_configs=rule_config.get("parameter_builders"),
            data_context=self._data_context,
        )
        expectation_configuration_builders: List[
            ExpectationConfigurationBuilder
        ] = init_rule_expectation_configuration_builders(
            expectation_configuration_builder_configs=rule_config.get(
                "expectation_configuration_builders"
            ),
            data_context=self._data_context,
        )

        # Compile previous steps and package into a Rule object
        return Rule(
            name=rule_name,
            variables=variables,
            domain_builder=domain_builder,
            parameter_builders=parameter_builders,
            expectation_configuration_builders=expectation_configuration_builders,
        )

    @staticmethod
    def _init_rule_domain_builder(
        domain_builder_config: dict,
        data_context: Optional[AbstractDataContext] = None,
    ) -> DomainBuilder:
        domain_builder: DomainBuilder = instantiate_class_from_config(
            config=domain_builder_config,
            runtime_environment={"data_context": data_context},
            config_defaults={
                "module_name": "great_expectations.rule_based_profiler.domain_builder"
            },
        )

        return domain_builder

    @public_api
    @usage_statistics_enabled_method(
        event_name=UsageStatsEvents.RULE_BASED_PROFILER_RUN,
        args_payload_fn=get_profiler_run_usage_statistics,
    )
    def run(
        self,
        variables: Optional[Dict[str, Any]] = None,
        rules: Optional[Dict[str, Dict[str, Any]]] = None,
        batch_list: Optional[List[Batch]] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        runtime_configuration: Optional[dict] = None,
        reconciliation_directives: ReconciliationDirectives = DEFAULT_RECONCILATION_DIRECTIVES,
        variables_directives_list: Optional[
            List[RuntimeEnvironmentVariablesDirectives]
        ] = None,
        domain_type_directives_list: Optional[
            List[RuntimeEnvironmentDomainTypeDirectives]
        ] = None,
        comment: Optional[str] = None,
    ) -> RuleBasedProfilerResult:
        """Run the Rule-Based Profiler.

        Args:
            variables: Attribute name/value pairs (overrides), commonly-used in `Builder` objects.
            rules: A collection of rule configurations (overrides).
            batch_list: The batches of data supplied at runtime.
            batch_request: An explicit Batch Request used to supply data at runtime.
            runtime_configuration: Additional runtime settings (see `Validator.DEFAULT_RUNTIME_CONFIGURATION`).
            reconciliation_directives: Directives for how each rule component should be overwritten.
            variables_directives_list: Additional override runtime variables directives (modify `BaseRuleBasedProfiler`).
            domain_type_directives_list: Additional override runtime domain directives (modify `BaseRuleBasedProfiler`).
            comment: A citation for the Expectation Suite returned as part of the `RuleBasedProfilerResult`.

        Returns:
            A `RuleBasedProfilerResult` instance that contains the profiling output.
        """
        # Check to see if the user has disabled progress bars
        disable = False
        if self._data_context:
            progress_bars = self._data_context.progress_bars
            # If progress_bars are not present, assume we want them enabled
            if progress_bars is not None:
                if "globally" in progress_bars:
                    disable = not progress_bars["globally"]

                if "rule_based_profiler" in progress_bars:
                    disable = not progress_bars["rule_based_profiler"]

        effective_variables: Optional[
            ParameterContainer
        ] = self.reconcile_profiler_variables(
            variables=variables,
            reconciliation_strategy=reconciliation_directives.variables,
        )

        effective_rules: List[Rule] = self.reconcile_profiler_rules(
            rules=rules,
            reconciliation_directives=reconciliation_directives,
        )

        self._apply_runtime_environment(
            variables=effective_variables,
            rules=effective_rules,
            variables_directives_list=variables_directives_list,
            domain_type_directives_list=domain_type_directives_list,
        )

        rule: Rule
        effective_rules_configs: Optional[Dict[str, Dict[str, Any]]] = {
            rule.name: rule.to_json_dict() for rule in effective_rules
        }

        sys.stdout.write("\n")
        sys.stdout.flush()

        pbar_method: Callable = determine_progress_bar_method_by_environment()

        rule_state: RuleState
        rule: Rule
        for rule in pbar_method(
            effective_rules,
            desc="Generating Expectations:",
            disable=disable,
            position=0,
            leave=True,
            bar_format="{desc:25}{percentage:3.0f}%|{bar}{r_bar}",
        ):
            rule_state = rule.run(
                variables=effective_variables,
                batch_list=batch_list,
                batch_request=batch_request,
                runtime_configuration=runtime_configuration,
                reconciliation_directives=reconciliation_directives,
                rule_state=RuleState(),
            )
            self.rule_states.append(rule_state)

        return RuleBasedProfilerResult(
            fully_qualified_parameter_names_by_domain=self.get_fully_qualified_parameter_names_by_domain(),
            parameter_values_for_fully_qualified_parameter_names_by_domain=self.get_parameter_values_for_fully_qualified_parameter_names_by_domain(),
            expectation_configurations=list(
                filter(None, self.get_expectation_configurations())
            ),
            citation={
                "citation_date": convert_to_json_serializable(
                    data=datetime.datetime.now(datetime.timezone.utc)
                ),
                "comment": comment
                if comment
                else "Created by Rule-Based Profiler with the configuration included.",
                "profiler_config": {
                    "name": self.name,
                    "config_version": self.config_version,
                    "variables": convert_variables_to_dict(variables=self.variables),
                    "rules": effective_rules_configs,
                },
            },
            rule_domain_builder_execution_time={
                rule_state.rule.name: rule_state.rule_domain_builder_execution_time
                for rule_state in self.rule_states
            },
            rule_execution_time={
                rule_state.rule.name: rule_state.rule_execution_time
                for rule_state in self.rule_states
            },
            _usage_statistics_handler=self._usage_statistics_handler,
        )

    def get_expectation_configurations(self) -> List[ExpectationConfiguration]:
        """
        Returns:
            List of ExpectationConfiguration objects, accumulated from RuleState of every Rule executed.
        """
        expectation_configurations: List[ExpectationConfiguration] = []

        rule_state: RuleState
        rule_output: RuleOutput
        for rule_state in self.rule_states:
            rule_output = RuleOutput(rule_state=rule_state)
            expectation_configurations.extend(
                rule_output.get_expectation_configurations()
            )

        return expectation_configurations

    def get_fully_qualified_parameter_names_by_domain(self) -> Dict[Domain, List[str]]:
        """
        Returns:
            Dictionary of fully-qualified parameter names by Domain, accumulated from RuleState of every Rule executed.
        """
        fully_qualified_parameter_names_by_domain: Dict[Domain, List[str]] = {}

        rule_state: RuleState
        rule_output: RuleOutput
        for rule_state in self.rule_states:
            rule_output = RuleOutput(rule_state=rule_state)
            fully_qualified_parameter_names_by_domain.update(
                rule_output.get_fully_qualified_parameter_names_by_domain()
            )

        return fully_qualified_parameter_names_by_domain

    def get_fully_qualified_parameter_names_for_domain_id(
        self, domain_id: str
    ) -> List[str]:
        """
        Args:
            domain_id: ID of desired Domain object.

        Returns:
            List of fully-qualified parameter names for Domain with domain_id as specified, accumulated from RuleState of corresponding Rule executed.
        """
        rule_state: RuleState
        for rule_state in self.rule_states:
            domain: Domain = rule_state.get_domains_as_dict().get(domain_id)
            if domain is not None:
                rule_output = RuleOutput(rule_state=rule_state)
                return rule_output.get_fully_qualified_parameter_names_for_domain_id(
                    domain_id=domain_id
                )

    def get_parameter_values_for_fully_qualified_parameter_names_by_domain(
        self,
    ) -> Dict[Domain, Dict[str, ParameterNode]]:
        """
        Returns:
            Dictionaries of values for fully-qualified parameter names by Domain, accumulated from RuleState of every Rule executed.
        """
        values_for_fully_qualified_parameter_names_by_domain: Dict[
            Domain, Dict[str, ParameterNode]
        ] = {}

        rule_state: RuleState
        rule_output: RuleOutput
        for rule_state in self.rule_states:
            rule_output = RuleOutput(rule_state=rule_state)
            values_for_fully_qualified_parameter_names_by_domain.update(
                rule_output.get_parameter_values_for_fully_qualified_parameter_names_by_domain()
            )

        return values_for_fully_qualified_parameter_names_by_domain

    def get_parameter_values_for_fully_qualified_parameter_names_for_domain_id(
        self, domain_id: str
    ) -> Dict[str, ParameterNode]:
        """
        Args:
            domain_id: ID of desired Domain object.

        Returns:
            Dictionary of values for fully-qualified parameter names for Domain with domain_id as specified, accumulated from RuleState of corresponding Rule executed.
        """
        rule_state: RuleState
        for rule_state in self.rule_states:
            domain: Domain = rule_state.get_domains_as_dict().get(domain_id)
            if domain is not None:
                rule_output = RuleOutput(rule_state=rule_state)
                return rule_output.get_parameter_values_for_fully_qualified_parameter_names_for_domain_id(
                    domain_id=domain_id
                )

    def add_rule(self, rule: Rule) -> None:
        """
        Add Rule object to existing profiler object by reconciling profiler rules and updating _profiler_config.
        """
        rules_dict: Dict[str, Dict[str, Any]] = {
            rule.name: rule.to_json_dict(),
        }
        effective_rules: List[Rule] = self.reconcile_profiler_rules(
            rules=rules_dict,
            reconciliation_directives=ReconciliationDirectives(
                domain_builder=ReconciliationStrategy.UPDATE,
                parameter_builder=ReconciliationStrategy.UPDATE,
                expectation_configuration_builder=ReconciliationStrategy.UPDATE,
            ),
        )
        self.rules = effective_rules
        updated_rules: Optional[Dict[str, Dict[str, Any]]] = {
            rule.name: rule.to_json_dict() for rule in effective_rules
        }
        self._profiler_config.rules = updated_rules

    def reconcile_profiler_variables(
        self,
        variables: Optional[Dict[str, Any]] = None,
        reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.variables,
    ) -> Optional[ParameterContainer]:
        """
        Profiler "variables" reconciliation involves combining the variables, instantiated from Profiler configuration
        (e.g., stored in a YAML file managed by the Profiler store), with the variables overrides, provided at run time.

        The reconciliation logic for "variables" is of the "replace" nature: An override value complements the original
        on key "miss", and replaces the original on key "hit" (or "collision"), because "variables" is a unique member.

        :param variables: variables overrides, supplied in dictionary (configuration) form
        :param reconciliation_strategy: one of update, nested_update, or overwrite ways of reconciling overwrites
        :return: reconciled variables in their canonical ParameterContainer object form
        """
        effective_variables: ParameterContainer
        if variables and isinstance(variables, dict):
            variables_configs: dict = self._reconcile_profiler_variables_as_dict(
                variables=variables, reconciliation_strategy=reconciliation_strategy
            )
            effective_variables = build_parameter_container_for_variables(
                variables_configs=variables_configs
            )
        else:
            effective_variables = self.variables

        return effective_variables

    def _reconcile_profiler_variables_as_dict(
        self,
        variables: Optional[Dict[str, Any]],
        reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.variables,
    ) -> dict:
        if variables is None:
            variables = {}

        variables_configs: Optional[Dict[str, Any]] = convert_variables_to_dict(
            variables=self.variables
        )

        if reconciliation_strategy == ReconciliationStrategy.NESTED_UPDATE:
            variables_configs = nested_update(
                variables_configs,
                variables,
            )
        elif reconciliation_strategy == ReconciliationStrategy.REPLACE:
            variables_configs = variables
        elif reconciliation_strategy == ReconciliationStrategy.UPDATE:
            variables_configs.update(variables)

        return variables_configs

    def reconcile_profiler_rules(
        self,
        rules: Optional[Dict[str, Dict[str, Any]]] = None,
        reconciliation_directives: ReconciliationDirectives = DEFAULT_RECONCILATION_DIRECTIVES,
    ) -> List[Rule]:
        """
        Profiler "rules" reconciliation involves combining the rules, instantiated from Profiler configuration (e.g.,
        stored in a YAML file managed by the Profiler store), with the rules overrides, provided at run time.

        The reconciliation logic for "rules" is of the "procedural" nature:
        (1) Combine every rule override configuration with any instantiated rule into a reconciled configuration
        (2) Re-instantiate Rule objects from the reconciled rule configurations

        :param rules: rules overrides, supplied in dictionary (configuration) form for each rule name as the key
        :param reconciliation_directives directives for how each rule component should be overwritten
        :return: reconciled rules in their canonical List[Rule] object form
        """
        effective_rules: Dict[str, Rule] = self._reconcile_profiler_rules_as_dict(
            rules=rules,
            reconciliation_directives=reconciliation_directives,
        )
        return list(effective_rules.values())

    def _reconcile_profiler_rules_as_dict(
        self,
        rules: Optional[Dict[str, Dict[str, Any]]] = None,
        reconciliation_directives: ReconciliationDirectives = DEFAULT_RECONCILATION_DIRECTIVES,
    ) -> Dict[str, Rule]:
        if rules is None:
            rules = {}

        effective_rules: Dict[str, Rule] = self._get_rules_as_dict()

        rule_name: str
        rule_config: dict

        override_rule_configs: Dict[str, Dict[str, Any]] = {
            rule_name: RuleBasedProfiler._reconcile_rule_config(
                existing_rules=effective_rules,
                rule_name=rule_name,
                rule_config=rule_config,
                reconciliation_directives=reconciliation_directives,
            )
            for rule_name, rule_config in rules.items()
        }
        override_rules: Dict[str, Rule] = {
            rule_name: self._init_rule(rule_name=rule_name, rule_config=rule_config)
            for rule_name, rule_config in override_rule_configs.items()
        }
        effective_rules.update(override_rules)
        return effective_rules

    @classmethod
    def _reconcile_rule_config(
        cls,
        existing_rules: Dict[str, Rule],
        rule_name: str,
        rule_config: dict,
        reconciliation_directives: ReconciliationDirectives = DEFAULT_RECONCILATION_DIRECTIVES,
    ) -> Dict[str, Any]:
        """
        A "rule configuration" reconciliation is the process of combining the configuration of a single candidate
        override rule with at most one configuration corresponding to the list of rules instantiated from Profiler
        configuration (e.g., stored in a YAML file managed by the Profiler store).

        The reconciliation logic for "Rule configuration" employes the "by construction" principle:
        (1) Find a common configuration between the variables configuration, possibly supplied as part of the candiate
        override Rule configuration, and the variables configuration of an instantiated Rule
        (2) Find a common configuration between the domain builder configuration, possibly supplied as part of the
        candiate override Rule configuration, and the domain builder configuration of an instantiated Rule
        (3) Find common configurations between parameter builder configurations, possibly supplied as part of the
        candiate override Rule configuration, and the parameter builder configurations of an instantiated Rule
        (4) Find common configurations between expectation configuration builder configurations, possibly supplied as
        part of the candiate override Rule configuration, and the expectation configuration builder configurations of an
        instantiated Rule
        (5) Construct the reconciled Rule configuration dictionary using the formal Rule properties ("domain_builder",
        "parameter_builders", and "expectation_configuration_builders") as keys and their reconciled configuration
        dictionaries as values

        In order to insure successful instantiation of custom builder classes using "instantiate_class_from_config()",
        candidate builder override configurations are required to supply both "class_name" and "module_name" attributes.

        :param existing_rules: all currently instantiated rules represented as a dictionary, keyed by rule name
        :param rule_name: name of the override rule candidate
        :param rule_config: configuration of an override rule candidate, supplied in dictionary (configuration) form
        :param reconciliation_directives directives for how each rule component should be overwritten
        :return: reconciled rule configuration, returned in dictionary (configuration) form
        """
        effective_rule_config: Dict[str, Any]
        if rule_name in existing_rules:
            rule: Rule = existing_rules[rule_name]

            variables_config: dict = rule_config.get("variables", {})
            effective_variables: dict = reconcile_rule_variables(
                variables=rule.variables,
                variables_config=variables_config,
                reconciliation_strategy=reconciliation_directives.variables,
            )

            domain_builder_config: dict = rule_config.get("domain_builder", {})
            effective_domain_builder_config: dict = (
                cls._reconcile_rule_domain_builder_config(
                    domain_builder=rule.domain_builder,
                    domain_builder_config=domain_builder_config,
                    reconciliation_strategy=reconciliation_directives.domain_builder,
                )
            )

            parameter_builder_configs: List[dict] = rule_config.get(
                "parameter_builders", []
            )
            effective_parameter_builder_configs: Optional[
                List[dict]
            ] = cls._reconcile_rule_parameter_builder_configs(
                rule=rule,
                parameter_builder_configs=parameter_builder_configs,
                reconciliation_strategy=reconciliation_directives.parameter_builder,
            )

            expectation_configuration_builder_configs: List[dict] = rule_config.get(
                "expectation_configuration_builders", []
            )
            effective_expectation_configuration_builder_configs: List[
                dict
            ] = cls._reconcile_rule_expectation_configuration_builder_configs(
                rule=rule,
                expectation_configuration_builder_configs=expectation_configuration_builder_configs,
                reconciliation_strategy=reconciliation_directives.expectation_configuration_builder,
            )

            effective_rule_config = {
                "variables": effective_variables,
                "domain_builder": effective_domain_builder_config,
                "parameter_builders": effective_parameter_builder_configs,
                "expectation_configuration_builders": effective_expectation_configuration_builder_configs,
            }
        else:
            effective_rule_config = rule_config

        return effective_rule_config

    @staticmethod
    def _reconcile_rule_domain_builder_config(
        domain_builder: DomainBuilder,
        domain_builder_config: dict,
        reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.domain_builder,
    ) -> dict:
        """
        Rule "domain builder" reconciliation involves combining the domain builder, instantiated from Rule configuration
        (e.g., stored in a YAML file managed by the Profiler store), with the domain builder override, possibly supplied
        as part of the candiate override rule configuration.

        The reconciliation logic for "domain builder" is of the "replace" nature: An override value complements the
        original on key "miss", and replaces the original on key "hit" (or "collision"), because "domain builder" is a
        unique member for a Rule.

        :param domain_builder: existing domain builder of a Rule
        :param domain_builder_config: domain builder configuration override, supplied in dictionary (configuration) form
        :param reconciliation_strategy: one of update, nested_update, or overwrite ways of reconciling overwrites
        :return: reconciled domain builder configuration, returned in dictionary (configuration) form
        """
        domain_builder_as_dict: dict = domain_builder.to_json_dict()
        domain_builder_as_dict["class_name"] = domain_builder.__class__.__name__
        domain_builder_as_dict["module_name"] = domain_builder.__class__.__module__

        # Roundtrip through schema validation to add/or restore any missing fields.
        deserialized_config: DomainBuilderConfig = domainBuilderConfigSchema.load(
            domain_builder_as_dict
        )
        serialized_config: dict = deserialized_config.to_dict()

        effective_domain_builder_config: dict = serialized_config
        if domain_builder_config:
            _validate_builder_override_config(builder_config=domain_builder_config)
            if reconciliation_strategy == ReconciliationStrategy.NESTED_UPDATE:
                effective_domain_builder_config = nested_update(
                    effective_domain_builder_config,
                    domain_builder_config,
                )
            elif reconciliation_strategy == ReconciliationStrategy.REPLACE:
                effective_domain_builder_config = domain_builder_config
            elif reconciliation_strategy == ReconciliationStrategy.UPDATE:
                effective_domain_builder_config.update(domain_builder_config)

        return effective_domain_builder_config

    @staticmethod
    def _reconcile_rule_parameter_builder_configs(
        rule: Rule,
        parameter_builder_configs: List[dict],
        reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.parameter_builder,
    ) -> Optional[List[dict]]:
        """
        Rule "parameter builders" reconciliation involves combining the parameter builders, instantiated from Rule
        configuration (e.g., stored in a YAML file managed by the Profiler store), with the parameter builders
        overrides, possibly supplied as part of the candiate override rule configuration.

        The reconciliation logic for "parameter builders" is of the "upsert" nature: A candidate override parameter
        builder configuration contributes to the parameter builders list of the rule if the corresponding parameter
        builder name does not exist in the list of instantiated parameter builders of the rule; otherwise, once
        instnatiated, it replaces the configuration associated with the original parameter builder having the same name.

        :param rule: Profiler "rule", subject to parameter builder overrides
        :param parameter_builder_configs: parameter builder configuration overrides, supplied in dictionary (configuration) form
        :param reconciliation_strategy: one of update, nested_update, or overwrite ways of reconciling overwrites
        :return: reconciled parameter builder configuration, returned in dictionary (configuration) form
        """
        parameter_builder_config: dict
        for parameter_builder_config in parameter_builder_configs:
            _validate_builder_override_config(builder_config=parameter_builder_config)

        effective_parameter_builder_configs: Dict[str, dict] = {}

        current_parameter_builders: Dict[
            str, ParameterBuilder
        ] = rule._get_parameter_builders_as_dict()

        parameter_builder_name: str
        parameter_builder: ParameterBuilder
        parameter_builder_as_dict: dict
        for (
            parameter_builder_name,
            parameter_builder,
        ) in current_parameter_builders.items():
            parameter_builder_as_dict = parameter_builder.to_json_dict()
            parameter_builder_as_dict[
                "class_name"
            ] = parameter_builder.__class__.__name__
            parameter_builder_as_dict[
                "module_name"
            ] = parameter_builder.__class__.__module__

            # Roundtrip through schema validation to add/or restore any missing fields.
            deserialized_config: ParameterBuilderConfig = (
                parameterBuilderConfigSchema.load(parameter_builder_as_dict)
            )
            serialized_config: dict = deserialized_config.to_dict()

            effective_parameter_builder_configs[
                parameter_builder_name
            ] = serialized_config

        parameter_builder_configs_override: Dict[str, dict] = {
            parameter_builder_config["name"]: parameter_builder_config
            for parameter_builder_config in parameter_builder_configs
        }
        if reconciliation_strategy == ReconciliationStrategy.NESTED_UPDATE:
            effective_parameter_builder_configs = nested_update(
                effective_parameter_builder_configs,
                parameter_builder_configs_override,
                dedup=True,
            )
        elif reconciliation_strategy == ReconciliationStrategy.REPLACE:
            effective_parameter_builder_configs = parameter_builder_configs_override
        elif reconciliation_strategy == ReconciliationStrategy.UPDATE:
            effective_parameter_builder_configs.update(
                parameter_builder_configs_override
            )

        if not effective_parameter_builder_configs:
            return None

        return list(effective_parameter_builder_configs.values())

    @staticmethod
    def _reconcile_rule_expectation_configuration_builder_configs(
        rule: Rule,
        expectation_configuration_builder_configs: List[dict],
        reconciliation_strategy: ReconciliationStrategy = DEFAULT_RECONCILATION_DIRECTIVES.expectation_configuration_builder,
    ) -> List[dict]:
        """
        Rule "expectation configuration builders" reconciliation involves combining the expectation configuration builders, instantiated from Rule
        configuration (e.g., stored in a YAML file managed by the Profiler store), with the expectation configuration builders
        overrides, possibly supplied as part of the candiate override rule configuration.

        The reconciliation logic for "expectation configuration builders" is of the "upsert" nature: A candidate override expectation configuration
        builder configuration contributes to the expectation configuration builders list of the rule if the corresponding expectation configuration
        builder name does not exist in the list of instantiated expectation configuration builders of the rule; otherwise, once
        instnatiated, it replaces the configuration associated with the original expectation configuration builder having the same name.

        :param rule: Profiler "rule", subject to expectations configuration builder overrides
        :param expectation_configuration_builder_configs: expectation configuration builder configuration overrides, supplied in dictionary (configuration) form
        :param reconciliation_strategy: one of update, nested_update, or overwrite ways of reconciling overwrites
        :return: reconciled expectation configuration builder configuration, returned in dictionary (configuration) form
        """
        expectation_configuration_builder_config: dict
        for (
            expectation_configuration_builder_config
        ) in expectation_configuration_builder_configs:
            _validate_builder_override_config(
                builder_config=expectation_configuration_builder_config
            )

        effective_expectation_configuration_builder_configs: Dict[str, dict] = {}

        current_expectation_configuration_builders: Dict[
            str, ExpectationConfigurationBuilder
        ] = rule._get_expectation_configuration_builders_as_dict()

        expectation_configuration_builder_name: str
        expectation_configuration_builder: ExpectationConfigurationBuilder
        expectation_configuration_builder_as_dict: dict
        for (
            expectation_configuration_builder_name,
            expectation_configuration_builder,
        ) in current_expectation_configuration_builders.items():
            expectation_configuration_builder_as_dict = (
                expectation_configuration_builder.to_json_dict()
            )
            expectation_configuration_builder_as_dict[
                "class_name"
            ] = expectation_configuration_builder.__class__.__name__
            expectation_configuration_builder_as_dict[
                "module_name"
            ] = expectation_configuration_builder.__class__.__module__

            # Roundtrip through schema validation to add/or restore any missing fields.
            deserialized_config: ExpectationConfigurationBuilderConfig = (
                expectationConfigurationBuilderConfigSchema.load(
                    expectation_configuration_builder_as_dict
                )
            )
            serialized_config: dict = deserialized_config.to_dict()

            effective_expectation_configuration_builder_configs[
                expectation_configuration_builder_name
            ] = serialized_config

        expectation_configuration_builder_configs_override: Dict[str, dict] = {
            expectation_configuration_builder_config[
                "expectation_type"
            ]: expectation_configuration_builder_config
            for expectation_configuration_builder_config in expectation_configuration_builder_configs
        }
        if reconciliation_strategy == ReconciliationStrategy.NESTED_UPDATE:
            effective_expectation_configuration_builder_configs = nested_update(
                effective_expectation_configuration_builder_configs,
                expectation_configuration_builder_configs_override,
                dedup=True,
            )
        elif reconciliation_strategy == ReconciliationStrategy.REPLACE:
            effective_expectation_configuration_builder_configs = (
                expectation_configuration_builder_configs_override
            )
        elif reconciliation_strategy == ReconciliationStrategy.UPDATE:
            effective_expectation_configuration_builder_configs.update(
                expectation_configuration_builder_configs_override
            )

        if not effective_expectation_configuration_builder_configs:
            return []

        return list(effective_expectation_configuration_builder_configs.values())

    def _get_rules_as_dict(self) -> Dict[str, Rule]:
        rule: Rule
        return {rule.name: rule for rule in self.rules}

    # noinspection PyUnusedLocal
    def _apply_runtime_environment(
        self,
        variables: Optional[ParameterContainer] = None,
        rules: Optional[List[Rule]] = None,
        variables_directives_list: Optional[
            List[RuntimeEnvironmentVariablesDirectives]
        ] = None,
        domain_type_directives_list: Optional[
            List[RuntimeEnvironmentDomainTypeDirectives]
        ] = None,
    ) -> None:
        """
        variables: attribute name/value pairs, commonly-used in Builder objects, to modify using "runtime_environment"
        rules: name/(configuration-dictionary) to modify using "runtime_environment"
        variables_directives_list: additional/override runtime variables directives (modify "BaseRuleBasedProfiler")
        domain_type_directives_list: additional/override runtime domain directives (modify "BaseRuleBasedProfiler")
        """
        BaseRuleBasedProfiler._apply_variables_directives_runtime_environment(
            rules=rules,
            variables_directives_list=variables_directives_list,
        )
        BaseRuleBasedProfiler._apply_domain_type_directives_runtime_environment(
            rules=rules,
            domain_type_directives_list=domain_type_directives_list,
        )

        rule: Rule
        self._profiler_config.rules = {rule.name: rule.to_json_dict() for rule in rules}

    @staticmethod
    def _apply_variables_directives_runtime_environment(
        rules: Optional[List[Rule]] = None,
        variables_directives_list: Optional[
            List[RuntimeEnvironmentVariablesDirectives]
        ] = None,
    ) -> None:
        """
        rules: name/(configuration-dictionary) to modify using "runtime_environment"
        variables_directives_list: additional/override runtime variables directives (modify "BaseRuleBasedProfiler")
        """
        if rules is None:
            rules = []

        rule: Rule

        rule_names: List[str] = [rule.name for rule in rules]

        if variables_directives_list is None:
            variables_directives_list = []

        variables_directives_list = list(
            filter(
                lambda element: element.rule_name in rule_names,
                variables_directives_list,
            )
        )

        variables_directives: RuntimeEnvironmentVariablesDirectives

        rule_names = [
            variables_directives.rule_name
            for variables_directives in variables_directives_list
        ]

        rules = list(
            filter(
                lambda element: element.name in rule_names,
                rules,
            )
        )

        rules_as_dict: Dict[str, Rule] = {rule.name: rule for rule in rules}

        variables: Optional[Dict[str, Any]]
        rule_variables_configs: Optional[Dict[str, Any]]
        for variables_directives in variables_directives_list:
            variables = variables_directives.variables or {}
            rule = rules_as_dict[variables_directives.rule_name]
            rule_variables_configs = convert_variables_to_dict(variables=rule.variables)
            # noinspection PyTypeChecker
            variables = dict(
                filter(
                    lambda element: element[0] in rule_variables_configs,
                    variables.items(),
                )
            )
            rule_variables_configs.update(variables)
            rule.variables = build_parameter_container_for_variables(
                variables_configs=rule_variables_configs
            )

    @staticmethod
    def _apply_domain_type_directives_runtime_environment(
        rules: Optional[List[Rule]] = None,
        domain_type_directives_list: Optional[
            List[RuntimeEnvironmentDomainTypeDirectives]
        ] = None,
    ) -> None:
        """
        rules: name/(configuration-dictionary) to modify using "runtime_environment"
        domain_type_directives_list: additional/override runtime domain directives (modify "BaseRuleBasedProfiler")
        """
        if rules is None:
            rules = []

        if domain_type_directives_list is None:
            domain_type_directives_list = []

        domain_type_directives: RuntimeEnvironmentDomainTypeDirectives
        domain_rules: List[Rule]
        rule: Rule
        for domain_type_directives in domain_type_directives_list:
            domain_rules = [
                rule
                for rule in rules
                if rule.domain_builder.domain_type == domain_type_directives.domain_type
            ]
            property_key: str
            property_value: Any
            existing_property_value: Any
            for rule in domain_rules:
                for (
                    property_key,
                    property_value,
                ) in domain_type_directives.directives.items():
                    try:
                        # Insure that new directives augment (not eliminate) existing directives.
                        existing_property_value = getattr(
                            rule.domain_builder, property_key
                        )
                        property_value = BaseRuleBasedProfiler._get_effective_domain_builder_property_value(
                            dest_property_value=property_value,
                            source_property_value=existing_property_value,
                        )
                        setattr(rule.domain_builder, property_key, property_value)
                    except AttributeError:
                        # Skip every directive that is not defined property of "DomainBuilder" object of "domain_type".
                        pass

    @staticmethod
    def _get_effective_domain_builder_property_value(
        dest_property_value: Optional[Any] = None,
        source_property_value: Optional[Any] = None,
    ) -> Optional[Any]:
        # Property values of collections types must be unique (use set for "list"/"tuple" and "update" for dictionary).

        if isinstance(dest_property_value, list) and isinstance(
            source_property_value, list
        ):
            return list(set(dest_property_value + source_property_value))

        if isinstance(dest_property_value, tuple) and isinstance(
            source_property_value, tuple
        ):
            return tuple(set(dest_property_value + source_property_value))

        if isinstance(dest_property_value, dict) and isinstance(
            source_property_value, dict
        ):
            return dict(dest_property_value, **source_property_value)

        return dest_property_value

    @classmethod
    def run_profiler(
        cls,
        data_context: AbstractDataContext,
        profiler_store: ProfilerStore,
        batch_list: Optional[List[Batch]] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        name: Optional[str] = None,
        id: Optional[str] = None,
        variables: Optional[dict] = None,
        rules: Optional[dict] = None,
    ) -> RuleBasedProfilerResult:
        profiler: RuleBasedProfiler = cls.get_profiler(
            data_context=data_context,
            profiler_store=profiler_store,
            name=name,
            id=id,
        )

        return profiler.run(
            variables=variables,
            rules=rules,
            batch_list=batch_list,
            batch_request=batch_request,
            runtime_configuration=None,
            reconciliation_directives=DEFAULT_RECONCILATION_DIRECTIVES,
            variables_directives_list=None,
            domain_type_directives_list=None,
            comment=None,
        )

    @classmethod
    def run_profiler_on_data(
        cls,
        data_context: AbstractDataContext,
        profiler_store: ProfilerStore,
        batch_list: Optional[List[Batch]] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> RuleBasedProfilerResult:
        profiler: RuleBasedProfiler = cls.get_profiler(
            data_context=data_context,
            profiler_store=profiler_store,
            name=name,
            id=id,
        )

        rule: Rule
        rules: Dict[str, Dict[str, Any]] = {
            rule.name: rule.to_json_dict() for rule in profiler.rules
        }

        return profiler.run(
            variables=None,
            rules=rules,
            batch_list=batch_list,
            batch_request=batch_request,
            runtime_configuration=None,
            reconciliation_directives=DEFAULT_RECONCILATION_DIRECTIVES,
            variables_directives_list=None,
            domain_type_directives_list=None,
            comment=None,
        )

    @classmethod
    def add_profiler(
        cls,
        data_context: AbstractDataContext,
        profiler_store: ProfilerStore,
        name: str | None = None,
        id: str | None = None,
        config_version: float | None = None,
        rules: dict[str, dict] | None = None,
        variables: dict | None = None,
        profiler: RuleBasedProfiler | None = None,
    ) -> RuleBasedProfiler:
        return cls._persist_profiler(
            data_context=data_context,
            persistence_fn=profiler_store.add,
            name=name,
            id=id,
            config_version=config_version,
            rules=rules,
            variables=variables,
            profiler=profiler,
        )

    @classmethod
    def update_profiler(
        cls,
        profiler_store: ProfilerStore,
        data_context: AbstractDataContext,
        name: str | None = None,
        id: str | None = None,
        config_version: float | None = None,
        rules: dict[str, dict] | None = None,
        variables: dict | None = None,
        profiler: RuleBasedProfiler | None = None,
    ) -> RuleBasedProfiler:
        return cls._persist_profiler(
            data_context=data_context,
            persistence_fn=profiler_store.update,
            name=name,
            id=id,
            config_version=config_version,
            rules=rules,
            variables=variables,
            profiler=profiler,
        )

    @classmethod
    def add_or_update_profiler(
        cls,
        data_context: AbstractDataContext,
        profiler_store: ProfilerStore,
        name: str | None = None,
        id: str | None = None,
        config_version: float | None = None,
        rules: dict[str, dict] | None = None,
        variables: dict | None = None,
        profiler: RuleBasedProfiler | None = None,
    ) -> RuleBasedProfiler:
        return cls._persist_profiler(
            data_context=data_context,
            persistence_fn=profiler_store.add_or_update,
            name=name,
            id=id,
            config_version=config_version,
            rules=rules,
            variables=variables,
            profiler=profiler,
        )

    @classmethod
    def _persist_profiler(
        cls,
        data_context: AbstractDataContext,
        persistence_fn: Callable,
        name: str | None = None,
        id: str | None = None,
        config_version: float | None = None,
        rules: dict[str, dict] | None = None,
        variables: dict | None = None,
        profiler: RuleBasedProfiler | None = None,
    ) -> RuleBasedProfiler:
        config = RuleBasedProfiler._resolve_profiler_config_for_store(
            name=name,
            id=id,
            config_version=config_version,
            rules=rules,
            variables=variables,
            profiler=profiler,
        )

        if not RuleBasedProfiler._check_validity_of_batch_requests_in_config(
            config=config
        ):
            raise gx_exceptions.InvalidConfigError(
                "batch_data found in batch_request cannot be saved to ProfilerStore"
            )

        # Chetan - 20220204 - AbstractDataContext to be removed once it can be decoupled from RBP
        new_profiler: RuleBasedProfiler = instantiate_class_from_config(
            config=config.to_json_dict(),
            runtime_environment={
                "data_context": data_context,
            },
            config_defaults={
                "module_name": "great_expectations.rule_based_profiler",
                "class_name": "RuleBasedProfiler",
            },
        )

        key: Union[GXCloudIdentifier, ConfigurationIdentifier]

        from great_expectations.data_context.data_context.cloud_data_context import (
            CloudDataContext,
        )

        if isinstance(data_context, CloudDataContext):
            key = GXCloudIdentifier(resource_type=GXCloudRESTResource.PROFILER)
        else:
            key = ConfigurationIdentifier(
                configuration_key=config.name,
            )

        try:
            response = persistence_fn(key=key, value=config)
        except gx_exceptions.StoreBackendError as e:
            raise gx_exceptions.ProfilerError(
                f"{e.message}; could not persist profiler"
            ) from e

        if isinstance(response, GXCloudResourceRef):
            new_profiler.ge_cloud_id = response.cloud_id

        return new_profiler

    @staticmethod
    def _resolve_profiler_config_for_store(
        name: str | None = None,
        id: str | None = None,
        config_version: float | None = None,
        rules: dict[str, dict] | None = None,
        variables: dict | None = None,
        profiler: RuleBasedProfiler | None = None,
    ) -> RuleBasedProfilerConfig:
        if not (
            (profiler is None)
            ^ all(arg is None for arg in (name, config_version, rules))
        ):
            raise ValueError(
                "Must either pass in an existing profiler or individual constructor arguments (but not both)"
            )

        if profiler:
            config = profiler.config
            # Recognize any changes made to the runtime object
            config.config_version = profiler.config_version
            config.rules = {rule.name: rule.to_dict() for rule in profiler.rules}
            config.variables = convert_variables_to_dict(profiler.variables)
        else:
            config_data = {
                "name": name,
                "id": id,
                "config_version": config_version,
                "rules": rules,
                "variables": variables,
            }

            # Roundtrip through schema validation to remove any illegal fields add/or restore any missing fields.
            validated_config: dict = ruleBasedProfilerConfigSchema.load(config_data)
            profiler_config: dict = ruleBasedProfilerConfigSchema.dump(validated_config)
            profiler_config.pop("class_name")
            profiler_config.pop("module_name")

            config = RuleBasedProfilerConfig(**profiler_config)

        return config

    @staticmethod
    def _check_validity_of_batch_requests_in_config(
        config: RuleBasedProfilerConfig,
    ) -> bool:
        # Evaluate nested types in RuleConfig to parse out BatchRequests
        batch_requests: List[Union[BatchRequestBase, dict]] = []
        rule: dict
        for rule in config.rules.values():
            domain_builder: dict = rule["domain_builder"]
            if "batch_request" in domain_builder:
                batch_requests.append(domain_builder["batch_request"])

            parameter_builders: List[dict] = rule.get("parameter_builders", [])
            parameter_builder: dict
            for parameter_builder in parameter_builders:
                if "batch_request" in parameter_builder:
                    batch_requests.append(parameter_builder["batch_request"])

        # DataFrames shouldn't be saved to ProfilerStore
        batch_request: Union[BatchRequestBase, dict]
        for batch_request in batch_requests:
            if batch_request_contains_batch_data(batch_request=batch_request):
                return False

        return True

    @classmethod
    def get_profiler(
        cls,
        data_context: AbstractDataContext,
        profiler_store: ProfilerStore,
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> RuleBasedProfiler:
        key = cls._construct_profiler_key(name=name, id=id)
        try:
            profiler_config: RuleBasedProfilerConfig = profiler_store.get(key=key)
        except gx_exceptions.InvalidKeyError as exc_ik:
            config_id: Union[GXCloudIdentifier, ConfigurationIdentifier] = (
                key.configuration_key
                if isinstance(key, ConfigurationIdentifier)
                else key
            )
            raise gx_exceptions.ProfilerNotFoundError(
                message=f'Non-existent Profiler configuration named "{config_id}".\n\nDetails: {exc_ik}'
            )

        config: dict = profiler_config.to_json_dict()
        if name:
            config.update({"name": name})

        config = filter_properties_dict(properties=config, clean_falsy=True)

        profiler: RuleBasedProfiler = instantiate_class_from_config(
            config=config,
            runtime_environment={
                "data_context": data_context,
            },
            config_defaults={
                "module_name": "great_expectations.rule_based_profiler",
                "class_name": "RuleBasedProfiler",
            },
        )

        return profiler

    @classmethod
    def delete_profiler(
        cls,
        profiler_store: ProfilerStore,
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> None:
        key = cls._construct_profiler_key(name=name, id=id)

        try:
            profiler_store.remove_key(key=key)
        except (gx_exceptions.InvalidKeyError, KeyError) as exc_ik:
            config_id = (
                key.configuration_key
                if isinstance(key, ConfigurationIdentifier)
                else key
            )
            raise gx_exceptions.ProfilerNotFoundError(
                message=f'Non-existent Profiler configuration named "{config_id}".\n\nDetails: {exc_ik}'
            )

    @staticmethod
    def list_profilers(
        profiler_store: ProfilerStore,
        ge_cloud_mode: bool = False,
    ) -> List[str]:
        if ge_cloud_mode:
            return profiler_store.list_keys()
        return [x.configuration_key for x in profiler_store.list_keys()]

    @staticmethod
    def _construct_profiler_key(
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> ConfigurationIdentifier | GXCloudIdentifier:
        assert bool(name) ^ bool(id), "Must provide either name or id (but not both)"

        if id:
            return GXCloudIdentifier(
                resource_type=GXCloudRESTResource.PROFILER, cloud_id=id
            )
        return ConfigurationIdentifier(
            configuration_key=name,
        )

    def self_check(self, pretty_print: bool = True) -> dict:
        """
        Necessary to enable integration with `AbstractDataContext.test_yaml_config`
        Args:
            pretty_print: flag to turn on verbose output
        Returns:
            Dictionary that contains RuleBasedProfiler state
        """
        report_object: dict = {"config": self.config.to_json_dict()}

        if pretty_print:
            print(f"\nRuleBasedProfiler class name: {self.name}")

            if not self.variables:
                print(
                    'Your current RuleBasedProfiler configuration has an empty "variables" attribute. \
                    Please ensure you populate it if you\'d like to reference values in your "rules" attribute.'
                )

        return report_object

    @property
    def config(self) -> RuleBasedProfilerConfig:
        return self._profiler_config

    @property
    def name(self) -> str:
        return self._name

    @property
    def config_version(self) -> float:
        return self._config_version

    @property
    def variables(self) -> Optional[ParameterContainer]:
        # Returning a copy of the "self._variables" state variable in order to prevent write-before-read hazard.
        return copy.deepcopy(self._variables)

    @variables.setter
    def variables(self, value: Optional[ParameterContainer]) -> None:
        self._variables = value
        self.config.variables = convert_variables_to_dict(variables=value)

    @property
    def rules(self) -> List[Rule]:
        return self._rules

    @rules.setter
    def rules(self, value: List[Rule]) -> None:
        self._rules = value

    @property
    def rule_states(self) -> List[RuleState]:
        return self._rule_states

    @public_api
    def to_json_dict(self) -> dict:
        """Returns a JSON-serializable dict representation of this RuleBasedProfiler.

        Returns:
            A JSON-serializable dict representation of this RuleBasedProfiler.
        """
        variables_dict: Optional[Dict[str, Any]] = convert_variables_to_dict(
            variables=self.variables
        )

        rule: Rule
        serializeable_dict: dict = {
            "class_name": self.__class__.__name__,
            "module_name": self.__class__.__module__,
            "name": self.name,
            "config_version": self.config_version,
            "variables": variables_dict,
            "rules": [rule.to_json_dict() for rule in self.rules],
        }
        return serializeable_dict

    def __repr__(self) -> str:
        json_dict: dict = self.config.to_json_dict()
        return json.dumps(json_dict, indent=2)

    def __str__(self) -> str:
        return self.__repr__()


@public_api
class RuleBasedProfiler(BaseRuleBasedProfiler):
    """Create a `RuleBasedProfiler` to profile one or more batches of data.

    For each rule in the `rules` configuration, instantiate the following if
    available: a domain builder, a parameter builder, and a configuration builder.
    These will be used to define profiler computation patterns.

    Args:
        name: Give the Profiler a name.
        config_version: Specify the version of the Profiler to use (currently only `1.0` is supported).
        variables: Variables to be substituted within the rules.
        rules: A collection of rule configurations, each having its own `domain_builder`, `parameter_builders`, and `expectation_configuration_builders`.
        data_context: Define the full runtime environment (data access, etc.).

    Returns:
        A `RuleBasedProfiler` instance.

    --ge-feature-maturity-info--

        id: rule_based_profiler_overall
        title: Rule-Based Profiler
        icon:
        short_description: Configuration Driven Profiler
        description: Use YAML to configure a flexible Profiler engine, which will then generate an ExpectationSuite for a data set
        how_to_guide_url:
        maturity: Experimental
        maturity_details:
            api_stability: Low (instantiation of Profiler and the signature of the run() method will change)
            implementation_completeness: Moderate (some augmentation and/or growth in capabilities is to be expected)
            unit_test_coverage: High (but not complete -- additional unit tests will be added, commensurate with the upcoming new functionality)
            integration_infrastructure_test_coverage: N/A -> TBD
            documentation_completeness: Moderate
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: domain_builders
        title: Domain Builders
        icon:
        short_description: Configurable Domain builders for generating lists of ExpectationConfiguration objects
        description: Use YAML to build domains for ExpectationConfiguration generator (table, column, semantic types, etc.)
        how_to_guide_url:
        maturity: Experimental
        maturity_details:
            api_stability: Moderate
            implementation_completeness: Moderate (additional DomainBuilder classes will be developed)
            unit_test_coverage: High (but not complete -- additional unit tests will be added, commensurate with the upcoming new functionality)
            integration_infrastructure_test_coverage: N/A -> TBD
            documentation_completeness: Moderate
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: parameter_builders
        title: Parameter Builders
        icon:
        short_description: Configurable Parameter builders for generating parameters to be used by ExpectationConfigurationBuilder classes for generating lists of ExpectationConfiguration objects (e.g., as kwargs and meta arguments), corresponding to the Domain built by a DomainBuilder class
        description: Use YAML to configure single and multi batch based parameter computation modules for the use by ExpectationConfigurationBuilder classes
        how_to_guide_url:
        maturity: Experimental
        maturity_details:
            api_stability: Moderate
            implementation_completeness: Moderate (additional ParameterBuilder classes will be developed)
            unit_test_coverage: High (but not complete -- additional unit tests will be added, commensurate with the upcoming new functionality)
            integration_infrastructure_test_coverage: N/A -> TBD
            documentation_completeness: Moderate
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: expectation_configuration_builders
        title: ExpectationConfiguration Builders
        icon:
        short_description: Configurable ExpectationConfigurationBuilder classes for generating lists of ExpectationConfiguration objects (e.g., as kwargs and meta arguments), corresponding to the Domain built by a DomainBuilder class and using parameters, computed by ParameterBuilder classes
        description: Use YAML to configure ExpectationConfigurationBuilder classes, which emit lists of ExpectationConfiguration objects (e.g., as kwargs and meta arguments)
        how_to_guide_url:
        maturity: Experimental
        maturity_details:
            api_stability: Moderate
            implementation_completeness: Moderate (additional ExpectationConfigurationBuilder classes might be developed)
            unit_test_coverage: High (but not complete -- additional unit tests will be added, commensurate with the upcoming new functionality)
            integration_infrastructure_test_coverage: N/A -> TBD
            documentation_completeness: Moderate
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

    --ge-feature-maturity-info--
    """

    def __init__(
        self,
        name: str,
        config_version: float,
        variables: Optional[Dict[str, Any]] = None,
        rules: Optional[Dict[str, Dict[str, Any]]] = None,
        data_context: Optional[AbstractDataContext] = None,
        id: Optional[str] = None,
    ) -> None:
        """Initialize a RuleBasedProfiler."""

        profiler_config = RuleBasedProfilerConfig(
            name=name,
            id=id,
            config_version=config_version,
            variables=variables,
            rules=rules,
        )

        usage_statistics_handler: Optional[UsageStatisticsHandler] = None
        if data_context:
            usage_statistics_handler = data_context.usage_statistics_handler

        super().__init__(
            profiler_config=profiler_config,
            data_context=data_context,
            usage_statistics_handler=usage_statistics_handler,
        )


def _validate_builder_override_config(builder_config: dict) -> None:
    """
    In order to insure successful instantiation of custom builder classes using "instantiate_class_from_config()",
    candidate builder override configurations are required to supply both "class_name" and "module_name" attributes.

    :param builder_config: candidate builder override configuration
    :raises: ProfilerConfigurationError
    """
    if not all(
        (
            isinstance(builder_config, dict),
            "class_name" in builder_config,
            "module_name" in builder_config,
        )
    ):
        raise gx_exceptions.ProfilerConfigurationError(
            'Both "class_name" and "module_name" must be specified.'
        )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import TypeGuard

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.domain import (
    InferredSemanticDomainType,
    SemanticDomainTypes,
)
from great_expectations.core.profiler_types_mapping import ProfilerTypeMapping
from great_expectations.rule_based_profiler.semantic_type_filter import (
    SemanticTypeFilter,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.validator.validator import Validator

T = TypeVar("T")


def _is_sequence_of(sequence: Sequence, type_: Type[T]) -> TypeGuard[Sequence[T]]:
    return all(isinstance(x, type_) for x in sequence)


class SimpleSemanticTypeFilter(SemanticTypeFilter):
    """
    This class provides default implementation methods, any of which can be overwritten with different mechanisms.
    """

    def __init__(
        self,
        batch_ids: Optional[List[str]] = None,
        validator: Optional[Validator] = None,
        column_names: Optional[List[str]] = None,
    ) -> None:
        self._build_table_column_name_to_inferred_semantic_domain_type_map(
            batch_ids=batch_ids,  # type: ignore[arg-type] # could be None
            validator=validator,  # type: ignore[arg-type] # could be None
            column_names=column_names,  # type: ignore[arg-type] # could be None
        )

    @property
    def table_column_name_to_inferred_semantic_domain_type_map(
        self,
    ) -> Dict[str, SemanticDomainTypes]:
        return self._table_column_name_to_inferred_semantic_domain_type_map  # type: ignore[return-value] # could be None

    def parse_semantic_domain_type_argument(
        self,
        semantic_types: Optional[
            Union[str, SemanticDomainTypes, List[Union[str, SemanticDomainTypes]]]
        ] = None,
    ) -> List[SemanticDomainTypes]:
        if semantic_types is None:
            return []

        semantic_type: Union[str, SemanticDomainTypes]
        if isinstance(semantic_types, str):
            semantic_types = semantic_types.lower()
            return [
                SemanticDomainTypes(semantic_type) for semantic_type in [semantic_types]
            ]

        if isinstance(semantic_types, SemanticDomainTypes):
            return [semantic_type for semantic_type in [semantic_types]]

        if isinstance(semantic_types, list):
            if _is_sequence_of(semantic_types, str):
                semantic_types = [
                    semantic_type.lower() for semantic_type in semantic_types
                ]
                return [
                    SemanticDomainTypes(semantic_type)
                    for semantic_type in semantic_types
                ]

            if _is_sequence_of(semantic_types, SemanticDomainTypes):
                return [semantic_type for semantic_type in semantic_types]

            raise ValueError(
                "All elements in semantic_types list must be either of str or SemanticDomainTypes type."
            )

        raise ValueError("Unrecognized semantic_types directive.")

    def _build_table_column_name_to_inferred_semantic_domain_type_map(
        self,
        batch_ids: List[str],
        validator: Validator,
        column_names: List[str],
    ) -> None:
        column_types_dict_list: List[Dict[str, Any]] = validator.get_metric(
            metric=MetricConfiguration(
                metric_name="table.column_types",
                metric_domain_kwargs={
                    "batch_id": batch_ids[-1],  # active_batch_id
                },
                metric_value_kwargs={
                    "include_nested": True,
                },
            )
        )

        column_name: str
        self._table_column_name_to_inferred_semantic_domain_type_map = {
            column_name: self._infer_semantic_domain_type_from_table_column_type(
                column_types_dict_list=column_types_dict_list,
                column_name=column_name,
            ).semantic_domain_type
            for column_name in column_names
        }

    @staticmethod
    def _infer_semantic_domain_type_from_table_column_type(
        column_types_dict_list: List[Dict[str, Any]],
        column_name: str,
    ) -> InferredSemanticDomainType:
        # Note: As of Python 3.8, specifying argument type in Lambda functions is not supported by Lambda syntax.
        column_types_dict_list = list(
            filter(
                lambda column_type_dict: column_name == column_type_dict["name"],
                column_types_dict_list,
            )
        )
        if len(column_types_dict_list) != 1:
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""Error: {len(column_types_dict_list)} columns were found while obtaining semantic type \
    information.  Please ensure that the specified column name refers to exactly one column.
    """
            )

        column_type: str = str(column_types_dict_list[0]["type"]).upper()

        semantic_column_type: SemanticDomainTypes
        if any(
            map(
                lambda type_name: column_type.startswith(type_name.upper()),
                ProfilerTypeMapping.INT_TYPE_NAMES
                + ProfilerTypeMapping.FLOAT_TYPE_NAMES,
            )
        ):
            semantic_column_type = SemanticDomainTypes.NUMERIC
        elif column_type in {
            type_name.upper() for type_name in ProfilerTypeMapping.STRING_TYPE_NAMES
        }:
            semantic_column_type = SemanticDomainTypes.TEXT
        elif column_type in {
            type_name.upper() for type_name in ProfilerTypeMapping.BOOLEAN_TYPE_NAMES
        }:
            semantic_column_type = SemanticDomainTypes.LOGIC
        elif any(
            map(
                lambda type_name: column_type.startswith(type_name.upper()),
                ProfilerTypeMapping.DATETIME_TYPE_NAMES,
            )
        ):
            semantic_column_type = SemanticDomainTypes.DATETIME
        elif column_type in {
            type_name.upper() for type_name in ProfilerTypeMapping.BINARY_TYPE_NAMES
        }:
            semantic_column_type = SemanticDomainTypes.BINARY
        elif column_type in {
            type_name.upper() for type_name in ProfilerTypeMapping.CURRENCY_TYPE_NAMES
        }:
            semantic_column_type = SemanticDomainTypes.CURRENCY
        elif column_type in {
            type_name.upper() for type_name in ProfilerTypeMapping.IDENTIFIER_TYPE_NAMES
        }:
            semantic_column_type = SemanticDomainTypes.IDENTIFIER
        elif column_type in (
            {
                type_name.upper()
                for type_name in ProfilerTypeMapping.MISCELLANEOUS_TYPE_NAMES
            }
            | {type_name.upper() for type_name in ProfilerTypeMapping.RECORD_TYPE_NAMES}
        ):
            semantic_column_type = SemanticDomainTypes.MISCELLANEOUS
        else:
            semantic_column_type = SemanticDomainTypes.UNKNOWN

        inferred_semantic_column_type: InferredSemanticDomainType = (
            InferredSemanticDomainType(
                semantic_domain_type=semantic_column_type,
                details={
                    "algorithm_type": "deterministic",
                    "mechanism": "lookup_table",
                    "source": "great_expectations.profile.base.ProfilerTypeMapping",
                },
            )
        )

        return inferred_semantic_column_type



from __future__ import annotations

import logging
from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, List, Optional

from great_expectations.core.batch import (
    Batch,
    BatchDataType,
    BatchDefinition,
    BatchMarkers,
    _get_x_batch_request_class,
)

if TYPE_CHECKING:
    from great_expectations.core.id_dict import BatchSpec
    from great_expectations.execution_engine import ExecutionEngine

logger = logging.getLogger(__name__)
logging.captureWarnings(True)


class BatchManager:
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        batch_list: Optional[List[Batch]] = None,
    ) -> None:
        """
        Args:
            execution_engine: The ExecutionEngine to be used to access cache of loaded Batch objects.
            batch_list: List of Batch objects available from external source (default is None).
        """
        self._execution_engine: ExecutionEngine = execution_engine

        self._active_batch_id: Optional[str] = None
        self._active_batch_data_id: Optional[str] = None

        self._batch_cache: Dict[str, Batch] = OrderedDict()
        self._batch_data_cache: Dict[str, BatchDataType] = {}

        if batch_list:
            self.load_batch_list(batch_list=batch_list)

    @property
    def batch_data_cache(self) -> Dict[str, BatchDataType]:
        """Dictionary of loaded BatchData objects."""
        return self._batch_data_cache

    @property
    def loaded_batch_ids(self) -> List[str]:
        """IDs of loaded BatchData objects."""
        return list(self._batch_data_cache.keys())

    @property
    def active_batch_data_id(self) -> Optional[str]:
        """
        The Batch ID for the default "BatchData" object.

        When a specific Batch is unavailable, then the data associated with the active_batch_data_id will be used.

        This is a "safety valve" provision.  If self._active_batch_data_id is unavailable (e.g., did not get set for
        some reason), then if there is exactly and unambiguously one loaded "BatchData" object, then it will play the
        role of the "active_batch_data_id", which is needed to compute a metric (by the particular ExecutionEngine).
        However, if there is more than one, then "active_batch_data_id" becomes ambiguous, and thus "None" is returned.
        """
        if self._active_batch_data_id is not None:
            return self._active_batch_data_id

        if len(self._batch_data_cache) == 1:
            return list(self._batch_data_cache.keys())[0]

        return None

    @property
    def active_batch_data(self) -> Optional[BatchDataType]:
        """The BatchData object from the currently-active Batch object."""
        if self.active_batch_data_id is None:
            return None

        return self._batch_data_cache.get(self.active_batch_data_id)

    @property
    def batch_cache(self) -> Dict[str, Batch]:
        """Getter for ordered dictionary (cache) of "Batch" objects in use (with batch_id as key)."""
        return self._batch_cache

    @property
    def active_batch_id(self) -> Optional[str]:
        """
        Getter for active Batch ID.

        Indeed, "active_batch_data_id" and "active_batch_id" can be different.  The former refers to the most recently
        loaded "BatchData" object, while the latter refers to the most recently requested "Batch" object.  In applicable
        situations, no new "BatchData" objects have been loaded; however, a new "Validator" object was instantiated with
        the list of "Batch" objects, each of whose BatchData has already been loaded (and cached).  Since BatchData IDs
        are from the same name space as Batch IDs, this helps avoid unnecessary loading of data from different backends.
        """
        active_batch_data_id: Optional[str] = self.active_batch_data_id
        if active_batch_data_id != self._active_batch_id:
            logger.warning(
                "ID of active Batch and ID of active loaded BatchData differ."
            )

        return self._active_batch_id

    @property
    def active_batch(self) -> Optional[Batch]:
        """Getter for active Batch"""
        active_batch_id: Optional[str] = self.active_batch_id
        batch: Optional[Batch] = (
            None if active_batch_id is None else self.batch_cache.get(active_batch_id)
        )
        return batch

    @property
    def active_batch_spec(self) -> Optional[BatchSpec]:
        """Getter for active batch's batch_spec"""
        if not self.active_batch:
            return None

        return self.active_batch.batch_spec

    @property
    def active_batch_markers(self) -> Optional[BatchMarkers]:
        """Getter for active batch's batch markers"""
        if not self.active_batch:
            return None

        return self.active_batch.batch_markers

    @property
    def active_batch_definition(self) -> Optional[BatchDefinition]:
        """Getter for the active batch's batch definition"""
        if not self.active_batch:
            return None

        return self.active_batch.batch_definition

    def reset_batch_cache(self) -> None:
        """Clears Batch cache"""
        self._batch_cache = OrderedDict()
        self._active_batch_id = None

    def load_batch_list(self, batch_list: List[Batch]) -> None:
        batch: Batch
        for batch in batch_list:
            try:
                assert isinstance(
                    batch, (Batch, _get_x_batch_request_class())
                ), "Batch objects provided to BatchManager must be formal Great Expectations Batch typed objects."
            except AssertionError as e:
                logger.error(str(e))

            self._execution_engine.load_batch_data(
                batch_id=batch.id, batch_data=batch.data
            )

            self._batch_cache[batch.id] = batch
            # We set the active_batch_id in each iteration of the loop to keep in sync with the active_batch_data_id
            # that has been loaded.  Hence, the final active_batch_id will be that of the final BatchData loaded.
            self._active_batch_id = batch.id

    def save_batch_data(self, batch_id: str, batch_data: BatchDataType) -> None:
        """
        Updates the data for the specified Batch in the cache
        """
        self._batch_data_cache[batch_id] = batch_data
        self._active_batch_data_id = batch_id



"""Rendering utility"""
from __future__ import annotations

import copy
import decimal
import locale
import re
import warnings
from typing import Any, List, Optional, Sequence, Tuple, Union

import pandas as pd

from great_expectations.core._docs_decorators import public_api
from great_expectations.data_context.types.resource_identifiers import (
    ValidationResultIdentifier,
)
from great_expectations.exceptions import RenderingError

DEFAULT_PRECISION = 4
# create a new context for this task
ctx = decimal.Context()
# Lowering precision from the system default (28) can allow additional control over display
ctx.prec = DEFAULT_PRECISION


@public_api
def num_to_str(
    f: float,
    precision: int = DEFAULT_PRECISION,
    use_locale: bool = False,
    no_scientific: bool = False,
) -> str:
    """Convert the given float to a string, centralizing standards for precision and decisions about scientific \
    notation. Adds an approximately equal sign in the event precision loss (e.g. rounding) has occurred.

    For more context, please review the following:
        https://stackoverflow.com/questions/38847690/convert-float-to-string-in-positional-format-without-scientific-notation-and-fa

    Args:
        f: The number to format.
        precision: The number of digits of precision to display (defaults to 4).
        use_locale: If True, use locale-specific formatting (e.g. adding thousands separators).
        no_scientific: If True, print all available digits of precision without scientific notation. This may insert
                       leading zeros before very small numbers, causing the resulting string to be longer than `precision`
                       characters.

    Returns:
        A string representation of the input float `f`, according to the desired parameters.
    """
    assert not (use_locale and no_scientific)
    if precision != DEFAULT_PRECISION:
        local_context = decimal.Context()
        local_context.prec = precision
    else:
        local_context = ctx
    # We cast to string; we want to avoid precision issues, but format everything as though it were a float.
    # So, if it's not already a float, we will append a decimal point to the string representation
    s = repr(f)
    if not isinstance(f, float):
        s += f"{locale.localeconv().get('decimal_point')}0"
    d = local_context.create_decimal(s)
    if no_scientific:
        result = format(d, "f")
    elif use_locale:
        result = format(d, "n")
    else:
        result = format(d, "g")
    if f != locale.atof(result):
        # result = '≈' + result
        #  ≈  # \u2248
        result = f"≈{result}"
    decimal_char = locale.localeconv().get("decimal_point")
    if "e" not in result and "E" not in result and decimal_char in result:
        result = result.rstrip("0").rstrip(decimal_char)
    return result


SUFFIXES = {1: "st", 2: "nd", 3: "rd"}


def ordinal(num):
    """Convert a number to ordinal"""
    # Taken from https://codereview.stackexchange.com/questions/41298/producing-ordinal-numbers/41301
    # Consider a library like num2word when internationalization comes
    if 10 <= num % 100 <= 20:
        suffix = "th"
    else:
        # the second parameter is a default.
        suffix = SUFFIXES.get(num % 10, "th")
    return str(num) + suffix


def resource_key_passes_run_name_filter(resource_key, run_name_filter):
    if type(resource_key) == ValidationResultIdentifier:
        run_name = resource_key.run_id.run_name
    else:
        raise TypeError(
            "run_name_filter filtering is only implemented for ValidationResultResources."
        )

    if run_name_filter.get("equals"):
        return run_name_filter.get("equals") == run_name
    elif run_name_filter.get("not_equals"):
        return run_name_filter.get("not_equals") != run_name
    elif run_name_filter.get("includes"):
        return run_name_filter.get("includes") in run_name
    elif run_name_filter.get("not_includes"):
        return run_name_filter.get("not_includes") not in run_name
    elif run_name_filter.get("matches_regex"):
        regex = run_name_filter.get("matches_regex")
        regex_match = re.search(regex, run_name)
        return False if regex_match is None else True
    elif run_name_filter.get("eq"):
        # deprecated-v0.11.9
        warnings.warn(
            "The 'eq' key will is deprecated as of v0.11.9 and will be removed in v0.16. Please use the renamed 'equals' key.",
            DeprecationWarning,
        )
        return run_name_filter.get("eq") == run_name
    elif run_name_filter.get("ne"):
        # deprecated-v0.11.9
        warnings.warn(
            "The 'ne' key will is deprecated as of v0.11.9 and will be removed in v0.16. Please use the renamed 'not_equals' key.",
            DeprecationWarning,
        )
        return run_name_filter.get("ne") != run_name


@public_api
def substitute_none_for_missing(
    kwargs: dict[str, Any], kwarg_list: Sequence[str]
) -> dict[str, Any]:
    """Utility function to plug Nones in when optional parameters are not specified in expectation kwargs.

    Args:
        kwargs: A dictionary of keyword arguments.
        kwargs_list: A list or sequence of strings representing all possible keyword parameters to a function.

    Returns:
        A copy of the original `kwargs` with missing keys from `kwarg_list` defaulted to `None`.

    ```python
    >>> result = substitute_none_for_missing(
    ...    kwargs={"a":1, "b":2},
    ...    kwarg_list=["c", "d"]
    ... )
    ... print(result)
    {"a":1, "b":2, "c": None, "d": None}
    ```

    This is helpful for standardizing the input objects for rendering functions.
    The alternative is lots of awkward `if "some_param" not in kwargs or kwargs["some_param"] == None:` clauses in renderers.
    """

    new_kwargs = copy.deepcopy(kwargs)
    for kwarg in kwarg_list:
        if kwarg not in new_kwargs:
            new_kwargs[kwarg] = None
    return new_kwargs


# NOTE: the method is pretty dirty
@public_api
def parse_row_condition_string_pandas_engine(
    condition_string: str, with_schema: bool = False
) -> tuple[str, dict]:
    """Parses the row condition string into a pandas engine compatible format.

    Args:
        condition_string: A pandas row condition string.
        with_schema: Return results in json schema format. Defaults to False.

    Returns:
        A tuple containing the template string and a `dict` of parameters.

    ```python
    >>> template_str, params = parse_row_condition_string_pandas_engine("Age in [0, 42]")
    >>> print(template_str)
    "if $row_condition__0"
    >>> params
    {"row_condition__0": "Age in [0, 42]"}
    ```

    """
    if len(condition_string) == 0:
        condition_string = "True"

    template_str = "if "
    params = {}

    condition_string = (
        condition_string.replace("&", " AND ")
        .replace(" and ", " AND ")
        .replace("|", " OR ")
        .replace(" or ", " OR ")
        .replace("~", " NOT ")
        .replace(" not ", " NOT ")
    )
    condition_string = " ".join(condition_string.split())

    # replace tuples of values by lists of values
    tuples_list = re.findall(r"\([^\(\)]*,[^\(\)]*\)", condition_string)
    for value_tuple in tuples_list:
        value_list = value_tuple.replace("(", "[").replace(")", "]")
        condition_string = condition_string.replace(value_tuple, value_list)

    # divide the whole condition into smaller parts
    conditions_list = re.split(r"AND|OR|NOT(?! in)|\(|\)", condition_string)
    conditions_list = [
        condition.strip()
        for condition in conditions_list
        if condition != "" and condition != " "
    ]

    for i, condition in enumerate(conditions_list):
        param_value = condition.replace(" NOT ", " not ")

        if with_schema:
            params[f"row_condition__{str(i)}"] = {
                "schema": {"type": "string"},
                "value": param_value,
            }
        else:
            params[f"row_condition__{str(i)}"] = param_value
            condition_string = condition_string.replace(
                condition, f"$row_condition__{str(i)}"
            )

    template_str += condition_string.lower()

    return template_str, params


@public_api
def handle_strict_min_max(params: dict) -> tuple[str, str]:
    """Utility function for the at least and at most conditions based on strictness.

    Args:
        params: Dictionary containing "strict_min" and "strict_max" booleans.

    Returns:
        Tuple of strings to use for the at least condition and the at most condition.
    """
    at_least_str = (
        "greater than"
        if params.get("strict_min") is True
        else "greater than or equal to"
    )
    at_most_str = (
        "less than" if params.get("strict_max") is True else "less than or equal to"
    )

    return at_least_str, at_most_str


def build_count_table(
    partial_unexpected_counts: List[dict], unexpected_count: int
) -> Tuple[List[str], List[List[Any]]]:
    """
    Used by _diagnostic_unexpected_table_renderer() method in Expectation to render
    Unexpected Counts table.

    Args:
        partial_unexpected_counts: list of dictionaries containing unexpected values and counts
        unexpected_count: total number of unexpected values. Used to build the header

    Returns:
        List of strings that will be rendered into DataDocs

    """
    table_rows: List[List[str]] = []
    total_count: int = 0

    for unexpected_count_dict in partial_unexpected_counts:
        value: Optional[Any] = unexpected_count_dict.get("value")
        count: Optional[int] = unexpected_count_dict.get("count")
        if count:
            total_count += count
        if value is not None and value != "":
            table_rows.append([value, count])
        elif value == "":
            table_rows.append(["EMPTY", count])
        else:
            table_rows.append(["null", count])

    # Check to see if we have *all* of the unexpected values accounted for. If so,
    # we show counts. If not then we show Sampled Unexpected Values
    if total_count == unexpected_count:
        header_row = ["Unexpected Value", "Count"]
    else:
        header_row = ["Sampled Unexpected Values", "Count"]
    return header_row, table_rows


def build_count_and_index_table(
    partial_unexpected_counts: List[dict],
    unexpected_index_list: List[dict],
    unexpected_count: int,
    unexpected_list: Optional[List[dict]] = None,
    unexpected_index_column_names: Optional[List[str]] = None,
) -> Tuple[List[str], List[List[Any]]]:
    """
        Used by _diagnostic_unexpected_table_renderer() method in Expectation to render
        Unexpected Counts and Indices table for ID/PK.
    Args:
        partial_unexpected_counts: list of dictionaries containing unexpected values and counts
        unexpected_index_list: list of dictionaries containing unexpected indices and their values
        unexpected_count: how many total unexpected values are there?
        unexpected_list: optional list of all unexpected values. Used with default Pandas unexpected
             indices (without defined id/pk columns)
        unexpected_index_column_names: list of unexpected_index_column_names

    Returns:
        List of strings that will be rendered into DataDocs

    """
    table_rows: List[List[str]] = []
    total_count: int = 0

    unexpected_index_df: pd.DataFrame = _convert_unexpected_indices_to_df(
        unexpected_index_list=unexpected_index_list,
        unexpected_index_column_names=unexpected_index_column_names,
        unexpected_list=unexpected_list,
        partial_unexpected_counts=partial_unexpected_counts,
    )
    if unexpected_index_df.empty:
        raise RenderingError(
            "GX ran into an issue while building count and index table for rendering. Please check your configuration."
        )

    # using default indices for Pandas
    if unexpected_index_column_names is None:
        unexpected_index_column_names = ["Index"]

    for index, row in unexpected_index_df.iterrows():

        row_list: List[Union[str, int]] = []

        unexpected_value = index
        count = int(row.Count)

        total_count += count

        if unexpected_value is not None and unexpected_value != "":
            row_list.append(unexpected_value)
            row_list.append(count)
        elif unexpected_value == "":
            row_list.append("EMPTY")
            row_list.append(count)
        else:
            unexpected_value = "null"
            row_list.append("null")
            row_list.append(count)
        for column_name in unexpected_index_column_names:
            row_list.append(str(row[column_name]))

        if len(row_list) > 0:
            table_rows.append(row_list)

    # Check to see if we have *all* of the unexpected values accounted for. If so,
    # we show counts. If not, we only show "sampled" unexpected values.
    if total_count == unexpected_count:
        header_row = ["Unexpected Value", "Count"]
        for column_name in unexpected_index_column_names:
            header_row.append(column_name)
    else:
        header_row = ["Sampled Unexpected Values", "Count"]
        for column_name in unexpected_index_column_names:
            header_row.append(column_name)
    return header_row, table_rows


def _convert_unexpected_indices_to_df(
    unexpected_index_list: List[Union[dict, int]],
    partial_unexpected_counts: List[dict],
    unexpected_index_column_names: Optional[List[str]] = None,
    unexpected_list: Optional[List[Any]] = None,
) -> pd.DataFrame:
    """
    Helper method to convert the list of unexpected indices into a DataFrame that can be used to
    display unexpected indices. domain_column_list (the list of column the Expectation is run on) is used
    as the index for the DataFrame, and the columns are the unexpected_index_column_names, or a default
    value in the case of Pandas, which provides default indices.

    In cases where the number of indices is too great (max 10 by default), the remaining values are
    truncated and the column contains "..." in their place.

                pk_1
    giraffe     3
    lion        4
    zebra       5 6 8 ...

    Args:
        unexpected_index_list : all unexpected values and indices
        partial_unexpected_counts : counts for unexpected values (max 20 by default)
        unexpected_index_column_names:  in the case of defining ID/PK columns
        unexpected_list: if we are using default Pandas output.
    Returns:
        pd.DataFrame that contains indices for unexpected values
    """
    if unexpected_index_column_names:
        # if we have defined unexpected_index_column_names for ID/PK
        unexpected_index_df: pd.DataFrame = pd.DataFrame(
            unexpected_index_list, dtype="string"
        )
        unexpected_index_df = unexpected_index_df.fillna(value="null")
        domain_column_name_list: List[str] = list(
            set(unexpected_index_list[0].keys()).difference(
                set(unexpected_index_column_names)
            )
        )
    elif unexpected_list:
        # if we are using default Pandas unexpected indices
        unexpected_index_df = pd.DataFrame(
            list(zip(unexpected_list, unexpected_index_list)),
            columns=["Value", "Index"],
            dtype="string",
        )
        unexpected_index_df = unexpected_index_df.fillna(value="null")
        domain_column_name_list: List[str] = ["Value"]
        unexpected_index_column_names: List[str] = ["Index"]
    else:
        return pd.DataFrame()

    # 1. groupby on domain columns, and turn id/pk into list
    all_unexpected_indices: pd.DataFrame = unexpected_index_df.groupby(
        domain_column_name_list
    ).agg(lambda y: list(y))

    # 2. add count
    col_to_count: str = unexpected_index_column_names[0]
    all_unexpected_indices["Count"] = all_unexpected_indices[col_to_count].apply(
        lambda x: len(x)
    )

    # 3. ensure index is a string
    all_unexpected_indices.index = all_unexpected_indices.index.map(str)

    # 4. truncate indices and replace with `...` if necessary
    for column in unexpected_index_column_names:
        all_unexpected_indices[column] = all_unexpected_indices[column].apply(
            lambda row: truncate_list_of_indices(row)
        )

    # 5. only keep the rows we are rendering
    filtered_unexpected_indices = all_unexpected_indices.head(
        len(partial_unexpected_counts)
    )
    return filtered_unexpected_indices


def truncate_list_of_indices(
    indices: List[Union[int, str]], max_index: int = 10
) -> str:
    """
    Lambda function used to take unexpected_indices and turn into a string that can be rendered in DataDocs.
    For lists that are greater than max_index, it will truncate the list and add a "..."

    Args:
        indices: indices to show
        max_index: Maximum number of indices to display before showing "..." in the column

    Returns:
        string of indices that are joined using ` `

    """
    if len(indices) > max_index:
        indices = indices[:max_index]
        indices.append("...")
    return ", ".join(map(str, indices))



import pytest


@pytest.fixture
def periodic_table_of_elements():
    # fmt: off
    data = [
        "Hydrogen", "Helium", "Lithium", "Beryllium", "Boron", "Carbon", "Nitrogen", "Oxygen", "Fluorine", "Neon",
        "Sodium", "Magnesium", "Aluminum", "Silicon", "Phosphorus", "Sulfur", "Chlorine", "Argon", "Potassium", "Calcium",
        "Scandium", "Titanium", "Vanadium", "Chromium", "Manganese", "Iron", "Cobalt", "Nickel", "Copper", "Zinc",
        "Gallium", "Germanium", "Arsenic", "Selenium", "Bromine", "Krypton", "Rubidium", "Strontium", "Yttrium", "Zirconium",
        "Niobium", "Molybdenum", "Technetium", "Ruthenium", "Rhodium", "Palladium", "Silver", "Cadmium", "Indium", "Tin",
        "Antimony", "Tellurium", "Iodine", "Xenon", "Cesium", "Barium", "Lanthanum", "Cerium", "Praseodymium", "Neodymium",
        "Promethium", "Samarium", "Europium", "Gadolinium", "Terbium", "Dysprosium", "Holmium", "Erbium", "Thulium", "Ytterbium",
        "Lutetium", "Hafnium", "Tantalum", "Tungsten", "Rhenium", "Osmium", "Iridium", "Platinum", "Gold", "Mercury",
        "Thallium", "Lead", "Bismuth", "Polonium", "Astatine", "Radon", "Francium", "Radium", "Actinium", "Thorium",
        "Protactinium", "Uranium", "Neptunium", "Plutonium", "Americium", "Curium", "Berkelium", "Californium", "Einsteinium", "Fermium",
        "Mendelevium", "Nobelium", "Lawrencium", "Rutherfordium", "Dubnium", "Seaborgium", "Bohrium", "Hassium", "Meitnerium", "Darmstadtium",
        "Roentgenium", "Copernicium", "Nihomium", "Flerovium", "Moscovium", "Livermorium", "Tennessine", "Oganesson",
    ]
    # fmt: on
    return data



# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py imports">
import pyspark
from ruamel import yaml

import great_expectations as gx
from great_expectations import DataContext
from great_expectations.core import ExpectationSuite
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.data_context.util import file_relative_path
from great_expectations.validator.validator import Validator

# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py get_context">
context: DataContext = gx.get_context()
# </snippet>
yaml = yaml.YAML(typ="safe")

spark_session: pyspark.sql.session.SparkSession = (
    gx.core.util.get_or_create_spark_application()
)

# create and load Expectation Suite
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py create_expectation_suite">
context.add_expectation_suite(
    expectation_suite_name="insert_your_expectation_suite_name_here"
)
# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py get_expectation_suite">
suite: ExpectationSuite = context.get_expectation_suite(
    expectation_suite_name="insert_your_expectation_suite_name_here"
)
# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py datasource_yaml">
datasource_yaml = f"""
name: my_spark_datasource
class_name: Datasource
module_name: great_expectations.datasource
execution_engine:
    module_name: great_expectations.execution_engine
    class_name: SparkDFExecutionEngine
data_connectors:
    my_runtime_data_connector:
        class_name: RuntimeDataConnector
        batch_identifiers:
            - some_key_maybe_pipeline_stage
            - some_other_key_maybe_airflow_run_id
"""
# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py add_datasource">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>
# RuntimeBatchRequest with batch_data as Spark Dataframe
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py path_to_file">
path_to_file: str = "some_path.csv"
# </snippet>
# Please note this override is only to provide good UX for docs and tests.
path_to_file: str = file_relative_path(
    __file__, "data/yellow_tripdata_sample_2019-01.csv"
)
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py pyspark_df">
df: pyspark.sql.dataframe.DataFrame = spark_session.read.csv(path_to_file)
# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py runtime_batch_request">
runtime_batch_request = RuntimeBatchRequest(
    datasource_name="my_spark_datasource",
    data_connector_name="my_runtime_data_connector",
    data_asset_name="insert_your_data_asset_name_here",
    runtime_parameters={"batch_data": df},
    batch_identifiers={
        "some_key_maybe_pipeline_stage": "ingestion step 1",
        "some_other_key_maybe_airflow_run_id": "run 18",
    },
)
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
path_to_file: str = file_relative_path(
    __file__, "data/yellow_tripdata_sample_2019-01.csv"
)

# RuntimeBatchRequest with path
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py runtime_batch_request_2">
runtime_batch_request = RuntimeBatchRequest(
    datasource_name="my_spark_datasource",
    data_connector_name="my_runtime_data_connector",
    data_asset_name="insert_your_data_asset_name_here",
    runtime_parameters={"path": path_to_file},
    batch_identifiers={
        "some_key_maybe_pipeline_stage": "ingestion step 1",
        "some_other_key_maybe_airflow_run_id": "run 18",
    },
)
# </snippet>
# Constructing Validator by passing in RuntimeBatchRequest
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py get_validator">
my_validator: Validator = context.get_validator(
    batch_request=runtime_batch_request,
    expectation_suite=suite,  # OR
    # expectation_suite_name=suite_name
)
# </snippet>
my_validator.head()

# Constructing Validator by passing in arguments
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py get_validator_2">
my_validator: Validator = context.get_validator(
    datasource_name="my_spark_datasource",
    data_connector_name="my_runtime_data_connector",
    data_asset_name="insert_your_data_asset_name_here",
    runtime_parameters={"path": path_to_file},
    batch_identifiers={
        "some_key_maybe_pipeline_stage": "ingestion step 1",
        "some_other_key_maybe_airflow_run_id": "run 18",
    },
    batch_spec_passthrough={
        "reader_method": "csv",
        "reader_options": {"delimiter": ",", "header": True},
    },
    expectation_suite=suite,  # OR
    # expectation_suite_name=suite_name
)
# </snippet>
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_create_a_batch_of_data_from_an_in_memory_spark_dataframe.py validator_head">
my_validator.head()
# </snippet>
